#****************************************************************************
#**
#**  File     :  /lua/sim/buffdefinition.lua
#**
#**  Copyright  2008 Gas Powered Games, Inc.  All rights reserved.
#****************************************************************************

import('/lua/sim/AdjacencyBuffs.lua')
import('/lua/sim/CheatBuffs.lua') # Buffs for AI Cheating

--[[
##################################################################
## BUFFS - Explanation of Blueprint
##################################################################

BuffBlueprint {
    Name = 'VeterancyHealth',       -- Specific name of buff, used to check if buff exists or has been applied
    DisplayName = 'VeterancyHealth',-- Not used
    BuffType = 'VETERANCYHEALTH',   -- 'Group' used when stacking buffs together. 
    Stacks = 'ALWAYS',              -- Stacks buffs together using this modifier (Always, Replaces, Ignores)
    Duration = -1,                  -- Specifies if this buff gets continually applied over time (eg Damage over time)
    Affects = {                     -- Table of effects that apply to the unit.
        MaxHealth = {               -- Matches up with 'atype' in Buffs.lua
            Add = 0,                -- Add the value
            Mult = 1.1,             -- Multiplies the value
        },
    },
}
--]]
##################################################################
## VETERANCY BUFFS - UNIT HEALTH
##################################################################

BuffBlueprint {
    Name = 'VeterancyHealth',
    DisplayName = 'VeterancyHealth',
    BuffType = 'VETERANCYHEALTH',
    Stacks = 'ALWAYS',
    Duration = -1,
    MaxLevel = 40,
    Affects = {
        MaxHealth = {
            Add = 0,
            Mult = 1.10,
        },
    },
}


##################################################################
## VETERANCY BUFFS - UNIT REGEN
##################################################################

BuffBlueprint {
    Name = 'VeterancyRegen',
    DisplayName = 'VeterancyRegen',
    BuffType = 'VETERANCYREGEN',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        SelfRepair = {
            Add = 2,
            Mult = 1,
        },
    },
}

##################################################################
## VETERANCY BUFFS - UNIT DAMAGE DPS
##################################################################

BuffBlueprint {
    Name = 'VeterancyDPS',
    DisplayName = 'VeterancyDPS',
    BuffType = 'VETERANCYDPS',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        DPS = {
            Add = 0,
            Mult = 1.06,
        },
    },
}

##################################################################
## VETERANCY BUFFS - UNIT DAMAGE
##################################################################

BuffBlueprint {
    Name = 'VeterancyDamage',
    DisplayName = 'VeterancyDamage',
    BuffType = 'VETERANCYDAMAGE',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        Damage = {
            Add = 0,
            Mult = 1.05,
        },
    },
}

##################################################################
## VETERANCY BUFFS - UNIT DAMAGE AOE
##################################################################

BuffBlueprint {
    MaxLevel = 20,
    Name = 'VeterancyDamageArea',
    DisplayName = 'VeterancyDamageArea',
    BuffType = 'VETERANCYDAMAGEAREA',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        DamageRadius = {
            Add = 0,
            Mult = 1.05,
        },
    },
}

##################################################################
## VETERANCY BUFFS - UNIT Weapon Range
##################################################################

BuffBlueprint {
    MaxLevel = 40,
    Name = 'VeterancyRange',
    DisplayName = 'VeterancyRange',
    BuffType = 'VETERANCYRANGE',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        MaxRadius = {
            Add = 0,
            Mult = 1.025,
        },
    },
}

##################################################################
## VETERANCY BUFFS - UNIT Speed
##################################################################

BuffBlueprint {
    MaxLevel = 20,
    Name = 'VeterancySpeed',
    DisplayName = 'VeterancySpeed',
    BuffType = 'VETERANCYSPEED',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        MoveMult = {
            Add = 0,
            Mult = 1.025,
        },
    },
}

BuffBlueprint {
    Name = 'VeterancyShield',
    DisplayName = 'VeterancyShield',
    BuffType = 'VETERANCYSHIELD',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        ShieldHP = {
            Add = 0,
            Mult = 1.1,
        },
        ShieldRegen = {
            Add = 0,
            Mult = 1.1,
        },
    },
}

BuffBlueprint {
    MaxLevel = 40,
    Name = 'VeterancyVision',
    DisplayName = 'VeterancyVision',
    BuffType = 'VETERANCYVISION',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        VisionRadius = {
            Add = 0,
            Mult = 1.025,
        },
    },
}

BuffBlueprint {
    MaxLevel = 40,
    Name = 'VeterancyOmniRadius',
    DisplayName = 'VeterancyOmniRadius',
    BuffType = 'VETERANCYOMNIRADIUS',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        OmniRadius = {
            Add = 0,
            Mult = 1.025,
        },
    },
}

BuffBlueprint {
    MaxLevel = 40,
    Name = 'VeterancyRadar',
    DisplayName = 'VeterancyRadar',
    BuffType = 'VETERANCYRADAR',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        RadarRadius = {
            Add = 0,
            Mult = 1.025,
        },
    },
}

BuffBlueprint {
    Name = 'VeterancyBuildRate',
    DisplayName = 'VeterancyBuildRate',
    BuffType = 'VETERANCYBUILDRATE',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        BuildRate = {
            Add = 2,
            Mult = 1,
        },
    },
}

BuffBlueprint {
    Name = 'VeterancyResourceProduction',
    DisplayName = 'VeterancyResourceProduction',
    BuffType = 'VeterancyResourceProduction',
    Stacks = 'ALWAYS',
    Duration = -1,
    Affects = {
        EnergyProductionBuf = {
            Add = 10,
            Mult = 1,
        },
        MassProductionBuf = {
            Add = 0.5,
            Mult = 1,
        },
    },
}

COMMANDVETERANCYPRODUCTION = {
                MaxLevel = 5,
                Stacks = 'ALWAYS',
                Duration = -1,
                Affects = {
                    EnergyProductionBuf = {
                        Add = 200,
                        Mult = 1,
                    },
                    MassProductionBuf = {
                        Add = 2,
                        Mult = 1,
                    },
                },
            },

BuffBlueprint {
    Name = 'VeterancyRateOfFire',
    DisplayName = 'VeterancyRateOfFire',
    BuffType = 'VETERANCYRATEOFFIRE',
    Stacks = 'ALWAYS',
    Duration = -1,
    MaxLevel = 50,
    Affects = {
        RateOfFireBuf = {
            Add = 0,
            Mult = 1.02,
            Ceil = 5,
        },
    },
}






__moduleinfo.auto_reload = true
