do
local OldProjectile = Projectile
Projectile = Class(OldProjectile) {

    ##################################################################
    #
    # Section to setup auto retargeting for any tracking projectile
    # This tracks incoming damage to a unit and redirects the projectile to another target
    # if the incoming damage is greater than the health of the unit. This is done because for some
    # reason, when a projectile loses its target, it *also* loses the ability to track, ever.
    #
    ##################################################################        
   
    OnCreate = function(self)
        OldProjectile.OnCreate(self)
        --if not (self:GetBlueprint().Physics.TrackTarget) then return end  -- This makes bombs vanish before impacting their target!!  Line below should fix.
        if not (self:GetBlueprint().Physics.TrackTarget) or self:GetBlueprint().Physics.StraightDownOrdinance or self:GetBlueprint().Physics.RealisticOrdinance then return end
        self:ForkThread(self.Retarget,self)
    end, 

    Retarget = function(self)
        #Wait here if target doesn't exist yet.
        while self and not self:GetTrackingTarget() do
            #LOG("||Waiting to get target")
            WaitSeconds(2)
        end
        ##LOG("|| Finding Missile Target")
        local target = self:GetTrackingTarget()
        if target and IsUnit(target) and not target:IsDead() and not target.IncomingDamage then
            target.IncomingDamage = 0
            #First projectile fired will stay locked on.
            target.IncomingDamage = target.IncomingDamage + self.DamageData.DamageAmount
            return
        end
        if not target.IncomingDamage then
            #LOG("|| Target missing incoming damage".. repr(target))
            target.IncomingDamage = 0
        end
        target.IncomingDamage = target.IncomingDamage + self.DamageData.DamageAmount
        
        while self do

            local unit = self:GetLauncher()

            if not unit then 
                ##LOG("||Launcher Destroyed")
                self:Destroy()
                return
            end  #Detonate when launcher is destroyed.

            if self:Get2DDistance(self:GetPosition(),unit:GetPosition()) > unit:GetBlueprint().Weapon[1].MaxRadius + 5 then
                ##LOG("||Out of range")
                self:Destroy()
                return
            end #Detonate if out of range
            
            #Check target for incoming damage
            local target = self:GetTrackingTarget() 
            if  target and IsUnit(target) and not target:IsDead() and target.IncomingDamage > ((target:GetHealth() + self.DamageData.DamageAmount)) then
                ##LOG("||Current Target has overkill - finding new target")
                #Current target has more than enough missiles aimed at it - lets find a new target
                local aiBrain = unit:GetAIBrain()
                local position = self:GetPosition()
                local targetlist = {}
                if EntityCategoryContains(categories.ANTINAVY,self) then
                    #LOG("||Finding Naval targets")
                    targetlist = aiBrain:GetUnitsAroundPoint( categories.NAVAL, position, 100 ,'Enemy')
                elseif  EntityCategoryContains(categories.ANTIAIR,self) then
                    #LOG("||Finding Air targets")
                    targetlist = aiBrain:GetUnitsAroundPoint( categories.AIR, position, 100 ,'Enemy')
                elseif EntityCategoryContains(categories.DIRECTFIRE ,self) then
                    #LOG("||Finding Ground targets")
                    targetlist = aiBrain:GetUnitsAroundPoint( categories.LAND, position, 100 ,'Enemy')
                else
                    #LOG("||No category found!")
                    #LOG("||" .. repr(self))
                end
                for k,v in targetlist do
                    if not v.IncomingDamage then
                        #First missile to aquire lock, locks onto target
                        #LOG("|| Locking onto first target")
                        v.IncomingDamage= self.DamageData.DamageAmount
                        target.IncomingDamage = target.IncomingDamage - self.DamageData.DamageAmount
                        self:SetNewTarget(v)
                        self:TrackTarget(true)
                        return
                        #If target is in range, and hasn't already been targetted by other missiles
                    elseif v and not v.IsDead() and self:Get2DDistance(v:GetPosition(),unit:GetPosition()) < unit:GetBlueprint().Weapon[1].MaxRadius 
                        and v.IncomingDamage < v:GetHealth() then
                        #LOG("|| Aquired new target")
                        v.IncomingDamage= (v.IncomingDamage) + self.DamageData.DamageAmount
                        target.IncomingDamage = target.IncomingDamage - self.DamageData.DamageAmount
                        self:SetNewTarget(v)
                        self:TrackTarget(true)
                        break
                    end
                end
            end
            WaitSeconds(1)
        end
    end,

    OnDestroy = function(self)
        local target = self:GetTrackingTarget()
        #Target is out of range or otherwise unreachable - remove my incoming damage
        if target and IsUnit(target) and not target:IsDead() and target.IncomingDamage then 
            target.IncomingDamage = target.IncomingDamage - self.DamageData.DamageAmount
            if target.IncomingDamage<=0 then
                target.IncomingDamage=0
            end
        end 
        OldProjectile.OnDestroy(self)
    end,
    
    
    DoDamage = function(self, instigator, damageData, targetEntity)
        local damage = damageData.DamageAmount
        if damage and damage > 0 then
            local radius = damageData.DamageRadius
            if radius and radius > 0 then
                if not damageData.DoTTime or damageData.DoTTime <= 0 then
                    self:DoTaperedDamage(instigator, self:GetPosition(), radius, damage, damageData.DamageType, damageData.DamageFriendly, damageData.DamageSelf or false, damageData.EdgeDamageRatio or 1)
                else
                    ForkThread(self.AreaDoTThread, self, instigator, self:GetPosition(), damageData.DoTPulses or 1, (damageData.DoTTime / (damageData.DoTPulses or 1)), radius, damage, damageData.DamageType, damageData.DamageFriendly, damageData.DamageSelf, damageData.EdgeDamageRatio)
                end
            #ONLY DO DAMAGE IF THERE IS DAMAGE DATA.  SOME PROJECTILE DO NOT DO DAMAGE WHEN THEY IMPACT.
            elseif damageData.DamageAmount and targetEntity then
                if not damageData.DoTTime or damageData.DoTTime <= 0 then
                    Damage(instigator, self:GetPosition(), targetEntity, damageData.DamageAmount, damageData.DamageType)
                else
                    ForkThread(DefaultDamage.UnitDoTThread, instigator, targetEntity, damageData.DoTPulses or 1, (damageData.DoTTime / (damageData.DoTPulses or 1)), damage, damageData.DamageType, damageData.DamageFriendly)
                end
            end
        end
    end,

    AreaDoTThread = function(self,instigator, position, pulses, pulseTime, radius, damage, damType, friendly, DamageSelf, EdgeDamageRatio)
        for i = 1, pulses do
            self:DoTaperedDamage(instigator, position, radius, damage, damType, friendly, DamageSelf or false, EdgeDamageRatio or 1)
            WaitSeconds(pulseTime)
        end
    end,


    DoTaperedDamage = function(self, instigator, position, damageRadius, damage, damageType, damageFriendly, damageSelf, EdgeDamageRatio)
        if damageRadius and damage then
            local Run1Step = false
            if not EdgeDamageRatio then
                Run1Step = true
            else
                if EdgeDamageRatio >= 1 then
                    Run1Step = true
                end
            end
            if damageRadius < 2 then
                Run1Step = true
            end
            if Run1Step then
                DamageArea(instigator, position, damageRadius, damage, damageType or 'Normal', damageFriendly or true, damageSelf or false)
            else
                local EdgeDamage = damage * EdgeDamageRatio
                local HalfWay = math.ceil(damageRadius / 2)
                local EdgeFloor = math.floor(damageRadius)
                local StepCount = EdgeFloor - HalfWay
                local DamageFalloff = damage - EdgeDamage
                local StepDamage = DamageFalloff / StepCount
                local ThisDamageRing = HalfWay
                for i=1,StepCount do
                    DamageArea(instigator, position, ThisDamageRing, StepDamage, damageType or 'Normal', damageFriendly or true, damageSelf or false)
                    ThisDamageRing = ThisDamageRing + 1
                end
                DamageArea(instigator, position, damageRadius, EdgeDamage, damageType or 'Normal', damageFriendly or true, damageSelf or false)
            end
        end
    end,

    PassDamageData = function(self, damageData)
        OldProjectile.PassDamageData(self,damageData)
        self.DamageData.EdgeDamageRatio = damageData.EdgeDamageRatio
    end,
    
    Get2DDistance = function(self,Pos1,Pos2)
        return VDist2(Pos1[1], Pos1[3], Pos2[1], Pos2[3])
    end,
} TypeClass = Projectile
end
