do
local OldUnit = Unit
Unit = Class(OldUnit) {

    vetEmitter = nil,
    MyACUs = {},

    RecloakCountdown = 0,         -- This is the public countdown timer for weapons to set when they fire and make a unit decloak.
    CloakManagerThread = 0,
    Stopped = true,
    CloakEffectEnabled = false,

    IntelSwitches = {},

    DamageMultiplier = 1,

    IsBeingTransported = false,

    FiniteAmmo = false,

    IsTeleportBeacon = false, 
    TeleportBeaconDetectionRadius = 30, 
    TeleportBeaconActive = false,
    TeleportBeaconIsBeingTransported = false,
    TeleportRecharging = false,
    InBeaconProximityTesting = false,
    NaturalTeleporter = false,
    HasTeleportUpgrade = false,
    IsTransportingTeleportBeacon = false,
    FreeTeleportRange = 0,
    TeleportEnergyCost = 0,
    TeleportRechargeTime = 0,

    TookDamageRecently = false,
    UnitSelfRepairRate = 0, --Free repairs that kick in after no combat for a while. This is the default, uses bp.RegenRate
    UnitBeingRepairedRate = 0, -- 'Cheap' repairs by engineers for units not under fire
    UnitRegenRate = 0, -- 'Damage Sponge' type regen
    
    UnitDoneBuilding = false,
    FreeRepairRate = 0,
    DamageSponge = false,
                                -- This is the global tweak value that can be used to raise or lower the
    FudgeFactor = 8,            -- amount of Free Repairs across the board.
                                -- 8 = 40dps for T1 engineers, 80 for T2 and 120 for T3

    DamageCredit = 0,           #Add a new field to track damage dealt
    KillCredit = 0,             #Kills calculated by damage
    DamageConsideredKill = 0,

    ##########################################################################
    ##   Startup Function
    ##      This is the place to fork all threads that *every* unit should fork after being built.
    ##      eg self:ForkThread(self.PingSonar,self)
    ##########################################################################
    OnStopBeingBuilt = function(self, builder, layer)
        OldUnit.OnStopBeingBuilt(self, builder, layer)
        -- This is the place to fork all threads that *every* unit should fork after being built.
        local bp = self:GetBlueprint()
        if bp.IsTeleportBeacon then
            self.IsTeleportBeacon = bp.IsTeleportBeacon
        end
        if bp.Physics.TeleportRange then
            self.FreeTeleportRange = bp.Physics.TeleportRange
        end
        if bp.Economy.TeleportEnergyMod then
            self.TeleportEnergyCost = bp.Economy.BuildCostEnergy * bp.Economy.TeleportEnergyMod
        end
        if bp.Physics.TeleportRechargeTime then
            self.TeleportRechargeTime = bp.Physics.TeleportRechargeTime
        end
        self.EnergyUsageRates = self:GetBlueprint().Economy.EnergyUsageRates
        if self.EnergyUsageRates then
            self:SetupEnergyUsage()
        end
        #Regen no longer depends on the 'damage sponge' tag
        self.UnitSelfRepairRate = self:GetBlueprint().Defense.SelfRepairRate or 0
        self.UnitRegenRate = self:GetBlueprint().Defense.RegenRate or 0
        self:ApplyRegen()
        
        self.UnitDoneBuilding = true
        
        #Seraphim get regen automatically.
        #if EntityCategoryContains(categories.SERAPHIM, self) then
        #    self.UnitRegenRate = self:GetBlueprint().Defense.RegenRate
        #end

        if self:GetBlueprint().Economy.CaptureRate then
            local CaptureRate = self:GetBlueprint().Economy.CaptureRate
            self:SetCaptureTimeMultiplier(1/CaptureRate)
        end

        self:ForkThread(self.AirStagingWatchThread, self)
        self:SetupAmmoBar()
        --self:InitializeIntel()

        self:GetACU()
        --if not _G.UnitLimitRemain then
        --    _G.UnitLimitRemain = {}
        --end

    end,

    ##########################################################################################
    ##  DEBUG SECTION OF CODE (Code appearing here can be deleted. Its just for bug hunting   
    ##########################################################################################


    ##########################################################################################
    ##  END OF DEBUG SECTION OF CODE                                                          
    ##########################################################################################


    ###################################################
    ##  Regen & Self Repair
    ###################################################
    
    DisableRegenAfterDamage = function(self)
        self.TookDamageRecently = true
        self:ApplyRegen()
        --LOG("|| Setting regen to zero")
        #self:SetRegenRate(0)
        WaitSeconds(10)
        self.TookDamageRecently = false
        self:ApplyRegen()
        --LOG("|| After 10 sec, Setting regen to " .. self.UnitSelfRepairRate + self.UnitBeingRepairedRate)
        #if not self:IsDead() then
        #    self:SetRegenRate(self.UnitSelfRepairRate + self.UnitBeingRepairedRate)
        #    self.TookDamageRecently = false
        #end
    end, 
    
    DoTakeDamage = function(self, instigator, amount, vector, damageType)
        if not self:IsDead() then
            --self:CreateVisionEntity(instigator:GetArmy())         -- Put this back in if bugfixed.  Sometimes it stays forever and does not die.  Also it should not be created for units under stealth protection.
            if self.DamageSponge == false then
                if self.DisableRegenAfterDamageThread then
                    self.DisableRegenAfterDamageThread:Destroy()
                    self.DisableRegenAfterDamageThread = nil
                end
                self.DisableRegenAfterDamageThread = self:ForkThread(self.DisableRegenAfterDamage, self)     -- This bit causes regen to pause in a fight.
            end
        end
        OldUnit.DoTakeDamage(self,instigator,amount,vector,damageType)
    end,
    
    ApplyRegen = function(self)
        #This should be the ONLY spot where SetRegenRate() is called from now on.
        #Everything else should set the appropriate variables and call this function.
        local rate
        if self.TookDamageRecently == false then
            --LOG("|| START: Setting Regen Rate to " .. unitBeingBuilt.UnitSelfRepairRate + unitBeingBuilt.UnitBeingRepairedRate)
            rate = self.UnitSelfRepairRate + self.UnitBeingRepairedRate + self.UnitRegenRate
        else
            rate = self.UnitRegenRate       
        end
        if not self:IsDead() then
            self:SetRegenRate(rate)
            self.Sync.RegenRate = rate
        end
    end,

    NOSTimeout = function(self)
        WaitSeconds(3)
        if self.NOSEffectBag then
            for k,v in self.NOSEffectBag do
                v:Destroy()
            end
            self.NOSEffectBag = {}      -- or nil?
        end
        self.NOSEffect = nil
    end,

    ##########################################################################
    ##   Unit Stat Manager
    ##        This section is overriding some of the 'Set' functions to track and 
    ##        allow units to modify their base values seperate from their 'temporary' values.
    ##########################################################################

    BaseSpeed = 0,
    SpeedAdjustmentPercent = 1,
    CurrentSpeedMultiplier = 1,

    AdjustBaseSpeed = function(self, NewSpeedPercent)
        #This takes the new speed and applies it to the unit
        self.SpeedAdjustmentPercent = NewSpeedPercent
        #LOG("Adjusting Base Speed to: ".. NewSpeedPercent)
        self:SetSpeedMult(self.CurrentSpeedMultiplier)
    end,

    SetSpeedMult = function(self, mult)
        local ActualSpeedPercentage = ((self.SpeedAdjustmentPercent) * mult)
        #LOG("Current Speed:" ..self.CurrentSpeedMultiplier .. "  New Speed: " .. ActualSpeedPercentage .. " SpeedMult: " .. mult)
        self.CurrentSpeedMultiplier = mult
        OldUnit.SetSpeedMult(self, ActualSpeedPercentage)
    end,
    
    # Feb 25 - Replaced with "ApplyRegen" function
    #Only syncs to the UI at the moment, but should also include
    #handling for damage sponge vs deferred regen.
    #SetRegenRate = function(self, rate)
    #    self.Sync.RegenRate = rate
    #    OldUnit.SetRegenRate(self, rate)
    # end,
    
    ###################################################
    #     Update Production Values
    #           This takes into account the increasing production values
    #           for hyrdros and mexes
    ##################################################

    UpdateProductionValues = function(self)
        local bpEcon = self:GetBlueprint().Economy
        if not bpEcon then return end
        self:SetProductionPerSecondEnergy((self.EnergyYield or bpEcon.ProductionPerSecondEnergy or 0) * (self.EnergyProdAdjMod or 1))
        self:SetProductionPerSecondMass((self.MassYield or bpEcon.ProductionPerSecondMass or 0) * (self.MassProdAdjMod or 1))
    end,

    ##########################################################################################
    ##  ENGINEERING
    ##  This makes repairs considerably cheaper IF they are done while not under fire.
    ##########################################################################################

    OnStartBuild = function(self, unitBeingBuilt, order)
		LOG("|| " .. order)
        if EntityCategoryContains(categories.ENGINEER, self) then
            LOG("|| Engineer: " .. self:GetBlueprint().Description)
            LOG("|| " .. order)
            if order != 'Upgrade' then
                if order == 'Repair' and unitBeingBuilt.UnitDoneBuilding == true then
                    if not unitBeingBuilt:IsUnitState('Upgrading') then
                        if self.BuildRateOverride and self.BuildRateOverride > 0 then
                            self.FreeRepairRate = self.BuildRateOverride
                            --LOG("||Using Build Rate Override")
                        else
                            self.FreeRepairRate = self:GetBlueprint().Economy.BuildRate
                            --LOG("||Using Build Rate")
                        end
                        if not Buff.HasBuff( unitBeingBuilt, 'UnitRepair'.. self:GetEntityId() ) then
                            BuffBlueprint {
                                Name = 'UnitRepair'.. self:GetEntityId(),
                                DisplayName = 'UnitRepair'.. self:GetEntityId(),
                                BuffType = 'Repair',
                                Stacks = 'ALWAYS',
                                Duration = -1,
                                Affects = {
                                    Repair = {
                                        Add =  self.FreeRepairRate,
                                        Mult = 1,
                                    },
                                },
                            }
                        end
                        Buff.ApplyBuff(unitBeingBuilt, 'UnitRepair'.. self:GetEntityId())
                        --unitBeingBuilt:ApplyRegen()
                    end
                end
            end
        end
        OldUnit.OnStartBuild(self,unitBeingBuilt,order)
    end,

    OnStopBuild = function(self, unitBeingBuilt)
        if unitBeingBuilt and not unitBeingBuilt:IsDead() then
            if self.FreeRepairRate > 0 then
                --unitBeingBuilt.UnitBeingRepairedRate = unitBeingBuilt.UnitBeingRepairedRate - (self.FreeRepairRate * self.FudgeFactor)
                if Buff.HasBuff( unitBeingBuilt, 'UnitRepair'.. self:GetEntityId() ) then
                    Buff.RemoveBuff( unitBeingBuilt, 'UnitRepair'.. self:GetEntityId() )
                end
                --unitBeingBuilt:ApplyRegen()
                #if unitBeingBuilt.TookDamageRecently == false then
                #    --LOG("|| DISENGAGING: Setting Regen Rate to " .. unitBeingBuilt.UnitSelfRepairRate + unitBeingBuilt.UnitBeingRepairedRate)
                #    unitBeingBuilt:SetRegenRate(unitBeingBuilt.UnitSelfRepairRate + unitBeingBuilt.UnitBeingRepairedRate)
                #end
                self.FreeRepairRate = 0
            end
            self:SetBuildTimeMultiplier(1)
        end
        OldUnit.OnStopBuild(self,unitBeingBuilt)
    end,

--[[
    CheckBuildRestriction = function(self, target_bp) 
        local brain = self:GetAIBrain() 
        local bpEcon = self:GetBlueprint().Economy

        local CustomCondition = true
        
        --local CustomCondition = brain:GetArmyStat("Units_Active_" .. target_bp.BlueprintId, 0.0).Value <= MaxBuiltNumber
        if self:CanBuild(target_bp.BlueprintId) and CustomCondition then 
            return true 
        else 
            return false 
        end 
    end,
--]]


    ##########################################################################################
    ##  CLOAKING
    ##########################################################################################

    CloakManager = function(self)
        local MainLoopDelay = 1      -- Recommend this not be any greater than 1.
        local JustDisabled = false
        local IntelBP = self:GetBlueprint().Intel
        self.bpDecloakRadius = IntelBP.DecloakRadius or 0
        self.bpDecloakRadiusWhileMoving = IntelBP.DecloakRadiusWhileMoving or 0
        self.bpDecloakWhileMoving = false
        if IntelBP.DecloakWhileMoving and IntelBP.DecloakWhileMoving == true then
            self.bpDecloakWhileMoving = true
        end
        self.bpCloakDisableAfterNoEnergyTimer = IntelBP.CloakDisableAfterNoEnergyTimer or 0
        while not self:IsDead() do
            if self:GetScriptBit('RULEUTC_CloakToggle') == false then
                if self.RecloakCountdown > 0 then                     -- If RecloakCountdown is above zero, remain decloaked until it is zero.
                    self:Decloak()
                    self.RecloakCountdown = self.RecloakCountdown - 1
                else
                    local DecloakRadius = 0
                    if self.Stopped then
                        DecloakRadius = self.bpDecloakRadius
                    else
                        DecloakRadius = self.bpDecloakRadiusWhileMoving
                    end
                    if self.bpDecloakWhileMoving == true then
                        if self.Stopped == false then
                            self:Decloak()
                            JustDisabled = true
                        end
                    end
                    if self:EnemyInSphere(DecloakRadius) then  -- Default is no decloak radius.
                        self:Decloak()
                    elseif self.UnitBeingBuilt then             -- Decloak when building
                        self:Decloak()
                    else
                        local aiBrain = self:GetAIBrain()
                        local EnergyStored = aiBrain:GetEconomyStored('ENERGY')
                        if EnergyStored < 10 and self.bpCloakDisableAfterNoEnergyTimer > MainLoopDelay then
                            self:Decloak()                                      -- Decloak if out of energy.
                        else
                            if JustDisabled == false then
                                if EnergyStored > 10 then
                                    self:Cloak()
                                end
                            end
                        end
                    end
                end
            end
            JustDisabled = false
            WaitSeconds(MainLoopDelay) 
        end
    end,

    Cloak = function(self)
        if not(self:IsIntelEnabled('Cloak')) then 
            self:EnableUnitIntel('Cloak')
            self:DestroyVeteranEmitter()
            if self.EnergySwitch then
                self.EnergySwitch.Cloak = true
                self:UpdateEnergyUsage()
            else
                self:SetMaintenanceConsumptionActive()
            end
        end
    end,

    Decloak = function(self)
        if self:IsIntelEnabled('Cloak') then 
            self:DisableUnitIntel('Cloak')
            self:CreateVeteranEmitter()
            if self.EnergySwitch then
                self.EnergySwitch.Cloak = false
                self:UpdateEnergyUsage()
            else
                self:SetMaintenanceConsumptionInactive()
            end
        end
    end,


    ##########################################################################################
    ##  TELEPORTING
    ##########################################################################################

    OnTeleportUnit = function(self, teleporter, location, orientation)
        local TeleportSuccessful = false
        if self.TeleportRecharging == false then
            local PhysicsBP = self:GetBlueprint().Physics
            local MyPosition = self:GetPosition() 
            if self:TeleportBeaconsInSphere(MyPosition) >= 1 then
                if self:TeleportBeaconsInSphere(location) >= 1 then
                    OldUnit.OnTeleportUnit(self, teleporter, location, orientation)
                    TeleportSuccessful = true
                end
            end
            if TeleportSuccessful == false and self.FreeTeleportRange > 0 then
                local Distance = VDist2(MyPosition[1],MyPosition[3],location[1],location[3])
                if Distance < self.FreeTeleportRange then
                    OldUnit.OnTeleportUnit(self, teleporter, location, orientation)
                    TeleportSuccessful = true
                end
            end
            if TeleportSuccessful == false and self.HasTeleportUpgrade == true and EntityCategoryContains(categories.COMMAND,self) then
                if self.TeleportUpgradeDisabled == false then
                    if self:TeleportBeaconsInSphere(location) >= 1 then
                        OldUnit.OnTeleportUnit(self, teleporter, location, orientation)
                        TeleportSuccessful = true
                    end
                end
            end
            if TeleportSuccessful == false and PhysicsBP and PhysicsBP.TeleportAnywhere then
                OldUnit.OnTeleportUnit(self, teleporter, location, orientation)
                TeleportSuccessful = true
            end
            if TeleportSuccessful == true then
                if self.TeleportRechargeTime > 0 then
                    self:ForkThread(self.TeleportRechargeTimer,self)
                else
                    if self.WhiteRabbitCountdown then
                        self.WhiteRabbitCountdown = self.WhiteRabbitCountdown + 20
                    end
                end
            else
                self:ForkThread(self.FlashFailedTeleportEffect,self)
            end
        end
    end,

    FlashFailedTeleportEffect = function(self)      -- Flash teleport effect to indicate failed teleport.
        self.TeleportRecharging = true
        self:ForkThread(self.CreateTeleportEffects,self)
        self:RemoveCommandCap('RULEUCC_Teleport')
        WaitSeconds(1.5)
        self:AddCommandCap('RULEUCC_Teleport')
        self.TeleportRecharging = false
        self:CleanupTeleportChargeEffects()
    end,

    TeleportBeaconsInSphere = function(self,pos) 
        local x1 = pos.x - self.TeleportBeaconDetectionRadius 
        local y1 = pos.y - self.TeleportBeaconDetectionRadius 
        local z1 = pos.z - self.TeleportBeaconDetectionRadius 
        local x2 = pos.x + self.TeleportBeaconDetectionRadius 
        local y2 = pos.y + self.TeleportBeaconDetectionRadius 
        local z2 = pos.z + self.TeleportBeaconDetectionRadius 
        local units = GetUnitsInRect( Rect(x1, z1, x2, z2) ) 
        
        if not units then 
            return false 
        end 
        
        local army = self:GetArmy() 
        local BeaconCount = 0
        for k,v in units do 
            if v.IsTeleportBeacon == true then
                if(army == v:GetArmy()) then 
                    local dist = self:Get2DDistance(pos,v:GetPosition()) 
                    if dist <= self.TeleportBeaconDetectionRadius then 
                        if v.TeleportBeaconActive == true then
                            BeaconCount = BeaconCount + 1
                        end
                    end 
                end 
            end
        end 
        return BeaconCount
    end, 

    TeleportRechargeTimer = function(self)
        self.TeleportRecharging = true
        WaitSeconds(1)
        self:ForkThread(self.CreateTeleportEffects,self)
        self:RemoveCommandCap('RULEUCC_Teleport')
        WaitSeconds(math.max(0,self.TeleportRechargeTime - 1))
        self:AddCommandCap('RULEUCC_Teleport')
        self.TeleportRecharging = false
        self:CleanupTeleportChargeEffects()
        if self.NaturalTeleporter == false then
            if self.BeaconProximityThread == nil then
                self.BeaconProximityThread = self:ForkThread(self.TestBeaconProximity,self)
            end
        end
    end,

    CreateTeleportEffects = function(self)
        local aiBrain = self:GetAIBrain()
        while not self:IsDead() and aiBrain:GetEconomyStored('ENERGY') < 10 do
            WaitSeconds(0.5)
        end
        local army = self:GetArmy()
        self.TeleportChargeBag = {}
        local Scale = self:GetEffectsScale()
        for k, v in EffectTemplate.GenericTeleportCharge01 do
            local fx = CreateAttachedEmitter(self, -1, army,v):ScaleEmitter(Scale)
            self.Trash:Add(fx)
            table.insert( self.TeleportChargeBag, fx)
        end
    end,

    -- The beacons start this thread when a non-natural teleporter enters the circle.
    -- The thread continues to check for the existence of TWO beacons in range, and manage the teleport button accordingly.
    -- Thread dies when unit is no longer in range of ANY beacons.  Then it is up to a new beacon to start it again.
    -- This thread also starts up when a non-natural teleporter's teleport recharge has completed, to see if 
    --   it is still in range of TWO beacons.
    TestBeaconProximity = function(self)
        self.InBeaconProximityTesting = true
        local TeleportEnabled = false
        local RequiredBeaconCount = 0
        if self.WhiteRabbitEnabled then
            RequiredBeaconCount = 0
        elseif self.HasTeleportUpgrade == true then
            if EntityCategoryContains(categories.COMMAND,self) then
                RequiredBeaconCount = 0
            else
                RequiredBeaconCount = 1
            end
        else
            if EntityCategoryContains(categories.NAVAL,self) then
                if EntityCategoryContains(categories.EXPERIMENTAL,self) then
                    RequiredBeaconCount = 4
                elseif EntityCategoryContains(categories.TECH3,self) then
                    RequiredBeaconCount = 4
                elseif EntityCategoryContains(categories.TECH2,self) then
                    RequiredBeaconCount = 3
                elseif EntityCategoryContains(categories.TECH1,self) then
                    RequiredBeaconCount = 2
                end
            else
                if EntityCategoryContains(categories.EXPERIMENTAL,self) then
                    RequiredBeaconCount = 4
                elseif EntityCategoryContains(categories.SUBCOMMANDER,self) then
                    RequiredBeaconCount = 1
                elseif EntityCategoryContains(categories.COMMAND,self) then
                    RequiredBeaconCount = 4
                elseif EntityCategoryContains(categories.TECH3,self) then
                    RequiredBeaconCount = 3
                elseif EntityCategoryContains(categories.TECH2,self) then
                    RequiredBeaconCount = 2
                elseif EntityCategoryContains(categories.TECH1,self) then
                    RequiredBeaconCount = 2
                end
            end
        end
        while (not self:IsDead()) and self.InBeaconProximityTesting == true do
            if self.IsTransportingTeleportBeacon == false then
                if self.TeleportRecharging == false then
                    local MyPosition = self:GetPosition() 
                    if self:TeleportBeaconsInSphere(MyPosition) >= RequiredBeaconCount then
                        if TeleportEnabled == false then
                            self:AddCommandCap('RULEUCC_Teleport')
                            TeleportEnabled = true
                        end
                    else
                        if TeleportEnabled == true then
                            self:RemoveCommandCap('RULEUCC_Teleport')
                            TeleportEnabled = false
                        end
                        if self:TeleportBeaconsInSphere(MyPosition) == 0 then
                            self.InBeaconProximityTesting = false
                        end
                    end
                end
            else
                if TeleportEnabled == true then
                    self:RemoveCommandCap('RULEUCC_Teleport')
                    TeleportEnabled = false
                end
            end
            WaitSeconds(1)
        end
        self.BeaconProximityThread:Destroy()
        self.BeaconProximityThread = nil
    end,

    -- This function exists here ONLY to allow for better control over teleport energy costs.
    -- (and to remove the call to PlayTeleportChargeEffects and CleanupTeleportChargeEffects
    InitiateTeleportThread = function(self, teleporter, location, orientation)
        local tbp = teleporter:GetBlueprint()
        local ubp = self:GetBlueprint()
        self.UnitBeingTeleported = self
        self:SetImmobile(true)
        self:PlayUnitSound('TeleportStart')
        self:PlayUnitAmbientSound('TeleportLoop')
        local bp = self:GetBlueprint().Economy
        local energyCost, time
        if bp then
            local mass = bp.BuildCostMass * (bp.TeleportMassMod or 0.01)
            local energy = self.TeleportEnergyCost
            energyCost = mass + energy
            time = energyCost * (bp.TeleportTimeMod or 0.01)
        end

        #LOG('*UNIT DEBUG: TELEPORTING, energyCost = ', repr(energyCost), ' time = ', repr(time))
        self.TeleportDrain = CreateEconomyEvent(self, energyCost or 100, 0, time or 5, self.UpdateTeleportProgress)

        # create teleport charge effect
        --self:PlayTeleportChargeEffects()

        WaitFor( self.TeleportDrain  ) # Perform fancy Teleportation FX here
        self.TeleportDrain = nil
        self:PlayTeleportOutEffects()
        --self:CleanupTeleportChargeEffects()

        WaitSeconds( 0.1 )

        #Teleport Sound
        #LOG('DRAIN')
        #self.PlayUnitSound ('TeleportDrain')
        if self.TeleportDrain then
            RemoveEconomyEvent(self, self.TeleportDrain )
            self.TeleportDrain = nil
        end

        self:SetWorkProgress(0.0)
        Warp(self, location, orientation)
        self:PlayTeleportInEffects()

        WaitSeconds( 0.1 ) # Perform cooldown Teleportation FX here
        #Landing Sound
        #LOG('DROP')
        self:StopUnitAmbientSound('TeleportLoop')
        self:PlayUnitSound('TeleportEnd')
        self:SetImmobile(false)
        self.UnitBeingTeleported = nil
        self.TeleportThread = nil
    end,



    ##########################################################################################
    ##  TRANSPORTS
    ##########################################################################################

    OnTransportAttach = function(self, attachBone, unit)
		LOG('Unit Attached')
		self:PlayUnitSound('Load')
        self:MarkWeaponsOnTransport(unit, true)
        if not EntityCategoryContains(categories.PODSTAGINGPLATFORM, self) then
            self:RequestRefreshUI()
        end
		if EntityCategoryContains(categories.AIRSTAGINGPLATFORM, self) then
			if not Buff.HasBuff( unit, 'UnitRepair'.. self:GetEntityId() ) then
				BuffBlueprint {
					Name = 'UnitRepair'.. self:GetEntityId(),
					DisplayName = 'UnitRepair'.. self:GetEntityId(),
					BuffType = 'Repair',
					Stacks = 'ALWAYS',
					Duration = -1,
					Affects = {
						Repair = {
							Add =  500,
							Mult = 1,
						},
					},
				}
			end
			Buff.ApplyBuff(unit, 'UnitRepair'.. self:GetEntityId())
		end
        unit.IsBeingTransported = true
        if unit.IsTeleportBeacon then
            if unit.IsTeleportBeacon == true then
                unit.TeleportBeaconIsBeingTransported = true
            end
        end
        self:RecalculateTransportSpeed()
		--OldUnit.OnTransportAttach(self, attachBone, unit)
    end,

    OnTransportDetach = function(self, attachBone, unit)
        OldUnit.OnTransportDetach(self, attachBone, unit) 
        unit.IsBeingTransported = false
        if unit.IsTeleportBeacon then
            if unit.IsTeleportBeacon == true then
                unit.TeleportBeaconIsBeingTransported = false
            end
        end
		if Buff.HasBuff( unit, 'UnitRepair'.. self:GetEntityId() ) then
			Buff.RemoveBuff( unit, 'UnitRepair'.. self:GetEntityId() )
		end
        self:RecalculateTransportSpeed()
        unit:TransportAnimation(-1)
    end,

    RecalculateTransportSpeed = function(self)
        local SpeedMultiplier = 1
        local bp = self:GetBlueprint().Transport
        local cargo = self:GetCargo()
        self.IsTransportingTeleportBeacon = false
        if table.getn(cargo) > 0 then
            for k, v in cargo do
                local TransportClass = v:GetBlueprint().Transport.TransportClass or 1
                if TransportClass == 1 then
                    SpeedMultiplier = SpeedMultiplier - (bp.Class1LoadSpeedPenalty or 0)
                elseif TransportClass == 2 then
                    SpeedMultiplier = SpeedMultiplier - (bp.Class2LoadSpeedPenalty or 0)
                elseif TransportClass == 3 then
                    SpeedMultiplier = SpeedMultiplier - (bp.Class3LoadSpeedPenalty or 0)
                end
                if v.IsTeleportBeacon then
                    if v.IsTeleportBeacon == true then
                        self.IsTransportingTeleportBeacon = true
                    end
                end
            end
        end
        --SpeedMultiplier = 1/((1/SpeedMultiplier)*2)
        self:SetSpeedMult(SpeedMultiplier)
        --self:SetAccMult(SpeedMultiplier)
        --self:SetTurnMult(SpeedMultiplier)
    end,

    OnLayerChange = function(self, new, old)      -- Set speed changes for layer transitions
        local PhysicsBP = self:GetBlueprint().Physics
        OldUnit.OnLayerChange(self, new, old)

        if new == 'Sub' then
            if PhysicsBP.LayerMods.Sub.Speed then
                self:SetSpeedMult(PhysicsBP.LayerMods.Sub.Speed)
            end
            if PhysicsBP.LayerMods.Sub.Acceleration then
                self:SetAccMult(PhysicsBP.LayerMods.Sub.Acceleration)
            end
            if PhysicsBP.LayerMods.Sub.TurnRate then
                self:SetTurnMult(PhysicsBP.LayerMods.Sub.TurnRate)
            end
        elseif new == 'Water' then
            if PhysicsBP.LayerMods.Water.Speed then
                self:SetSpeedMult(PhysicsBP.LayerMods.Water.Speed)
            end
            if PhysicsBP.LayerMods.Water.Acceleration then
                self:SetAccMult(PhysicsBP.LayerMods.Water.Acceleration)
            end
            if PhysicsBP.LayerMods.Water.TurnRate then
                self:SetTurnMult(PhysicsBP.LayerMods.Water.TurnRate)
            end
        elseif new == 'Land' then
            if PhysicsBP.LayerMods.Land.Speed then
                self:SetSpeedMult(PhysicsBP.LayerMods.Land.Speed)
            end
            if PhysicsBP.LayerMods.Land.Acceleration then
                self:SetAccMult(PhysicsBP.LayerMods.Land.Acceleration)
            end
            if PhysicsBP.LayerMods.Land.TurnRate then
                self:SetTurnMult(PhysicsBP.LayerMods.Land.TurnRate)
            end
        elseif new == 'Seabed' then
            if PhysicsBP.LayerMods.Land.Speed then
                self:SetSpeedMult(PhysicsBP.LayerMods.Seabed.Speed)
            end
            if PhysicsBP.LayerMods.Land.Acceleration then
                self:SetAccMult(PhysicsBP.LayerMods.Seabed.Acceleration)
            end
            if PhysicsBP.LayerMods.Land.TurnRate then
                self:SetTurnMult(PhysicsBP.LayerMods.Seabed.TurnRate)
            end
        end
    end, 

    DeathWeaponDamageThread = function(self,damageRadius, damage, damageType, damageFriendly, EdgeDamageRatio)
        if damageRadius and damage then
            local Run1Step = false
            if not EdgeDamageRatio then
                Run1Step = true
            else
                if EdgeDamageRatio >= 1 then
                    Run1Step = true
                end
            end
            if damageRadius < 2 then
                Run1Step = true
            end
            if Run1Step then
                DamageArea(self, self:GetPosition(), damageRadius, damage, damageType or 'Normal', damageFriendly or true)
            else
                local EdgeDamage = damage * EdgeDamageRatio
                local HalfWay = math.ceil(damageRadius / 2)
                local EdgeFloor = math.floor(damageRadius)
                local StepCount = EdgeFloor - HalfWay
                local DamageFalloff = damage - EdgeDamage
                local StepDamage = DamageFalloff / StepCount
                local ThisDamageRing = HalfWay
                for i=1,StepCount do 
                    DamageArea(self, self:GetPosition(), ThisDamageRing, StepDamage, damageType or 'Normal', damageFriendly or true)
                    ThisDamageRing = ThisDamageRing + 1
                end
                DamageArea(self, self:GetPosition(), damageRadius, EdgeDamage, damageType or 'Normal', damageFriendly or true)
            end
        end
    end,

    DoDeathWeapon = function(self)
        if self:IsBeingBuilt() then return end
        local bp = self:GetBlueprint()
        for k, v in bp.Weapon do
            if(v.Label == 'DeathWeapon') then
                if v.FireOnDeath == true then
                    self:SetWeaponEnabledByLabel('DeathWeapon', true)
                    self:GetWeaponByLabel('DeathWeapon'):Fire()
                else
                    self:ForkThread(self.DeathWeaponDamageThread, v.DamageRadius, v.Damage, v.DamageType, v.DamageFriendly, v.EdgeDamageRatio)
                end
                break
            end
        end
    end,

    SetupBuildBones = function(self)  -- This sets up the blueprint tag "BuildYawSpeed" so you can control the speed of construction aiming.
        local bp = self:GetBlueprint()
        if not bp.General.BuildBones or
           not bp.General.BuildBones.YawBone or
           not bp.General.BuildBones.PitchBone or
           not bp.General.BuildBones.AimBone then
           return
        end
        # Syntactical reference:
        #CreateBuilderArmController(unit,turretBone, [barrelBone], [aimBone])
        #BuilderArmManipulator:SetAimingArc(minHeading, maxHeading, headingMaxSlew, minPitch, maxPitch, pitchMaxSlew)
        self.BuildArmManipulator = CreateBuilderArmController(self, bp.General.BuildBones.YawBone or 0 , bp.General.BuildBones.PitchBone or 0, bp.General.BuildBones.AimBone or 0)
        local bp = self:GetBlueprint().Economy
        local BuildYawSpeed = bp.BuildYawSpeed or 360
        self.BuildArmManipulator:SetAimingArc(-180, 180, BuildYawSpeed, -90, 90, 360)
        self.BuildArmManipulator:SetPrecedence(5)
        if self.BuildingOpenAnimManip and self.BuildArmManipulator then
            self.BuildArmManipulator:Disable()
        end
        self.Trash:Add(self.BuildArmManipulator)
    end,

    OnDamage = function(self, instigator, amount, vector, damageType)
        amount = amount * self.DamageMultiplier 
        OldUnit.OnDamage(self, instigator, amount, vector, damageType)
    end, 

    SetArmoredState = function(self) 
        local ArmorMultiplier = self:GetBlueprint().Defense.ArmoredStateMultiplier or 1
        self.DamageMultiplier = ArmorMultiplier
    end, 

    SetUnarmoredState = function(self) 
        self.DamageMultiplier = 1 
    end,

    OnMotionHorzEventChange = function( self, new, old ) 
        OldUnit.OnMotionHorzEventChange(self, new, old) 
        if old == 'Stopped' then 
            self.Stopped = false
            local IntelBP = self:GetBlueprint().Intel
            if self.bpDecloakRadiusWhileMoving then
                self:SetIntelRadius('Sonar', self.bpDecloakRadiusWhileMoving)
            end
        elseif new == 'Stopped' then
            self.Stopped = true
            local IntelBP = self:GetBlueprint().Intel
            if self.bpDecloakRadius and self.bpDecloakRadiusWhileMoving then
                self:SetIntelRadius('Sonar', self.bpDecloakRadius)
            end
        end
    end,

    OnAnimCollision = function(self, bone, x, y, z)
        OldUnit.OnAnimCollision(self, bone, x, y, z)
        local layer = self:GetCurrentLayer()
        local bpTable = self:GetBlueprint().Display.MovementEffects

        if bpTable[layer].Footfall then
            bpTable = bpTable[layer].Footfall
            if bpTable.TreeFriendlyDamage then
                local bpDamage = bpTable.TreeFriendlyDamage
                local brain = self:GetAIBrain()
                local units = brain:GetUnitsAroundPoint(categories.VISIBLETORECON,self:GetPosition(bone),bpDamage.Radius,'Enemy')
                for k,v in units do
                    Damage(self,v:GetPosition(),v,bpDamage.Amount,'Normal')
                end
            end
        end
    end,


    ##########################################################################################
    ##  FINITE AMMO MANAGEMENT FOR AIRCRAFT
    ##########################################################################################

    SetupAmmoBar = function(self)
        if EntityCategoryContains(categories.AIR, self) then
            if EntityCategoryContains(categories.MOBILE, self) then
                local bpWeapon = self:GetBlueprint().Weapon
                if bpWeapon then
                    for k,v in bpWeapon do
                        if v.AmmoUsed then
                            self:SetFuelRatio(1)
                            self.FiniteAmmo = true
                            break
                        end
                    end
                end
            end
        end
    end,

    AirStagingWatchThread = function(self)
        local bp = self:GetBlueprint().AI
        local AirStagingRange = 0
        if bp.AirStagingRange then
            AirStagingRange = bp.AirStagingRange
        end
        WaitSeconds(5)
        if AirStagingRange > 0 then
            while not self:IsDead() do
                local pos = self:GetPosition() 
                local aiBrain = self:GetAIBrain()   
                local units = aiBrain:GetUnitsAroundPoint( categories.AIR, pos, AirStagingRange, 'Ally' )
                if units then
                    for k,v in units do
                        if v.FiniteAmmo == true then
                            if v:GetFuelRatio() < 1 then    -- If fuel is not 100%, make fuel 100%
                                v:SetFuelRatio(1)
                            end
                        end
                    end                
                end
                WaitSeconds(2)
            end        
        end
    end,

    OnRunOutOfFuel = function(self)
        self.HasFuel = false
        --self:DestroyTopSpeedEffects()
    end,


    ##########################################################################################
    ##  SHIELDS
    ##########################################################################################

    --[[
    EnableShield = function(self)
        self:SetScriptBit('RULEUTC_ShieldToggle', true)
        if self.MyShield then
            self.MyShield:Idle()
        end
    end,
    --]]

    ##########################################################################################
    ##  INTEL MANAGEMENT
    ##########################################################################################

    InitializeIntel = function(self)
        bpIntel = self:GetBlueprint().Intel
        if bpIntel.VisionRadius and bpIntel.VisionRadius > 0 then
            self.IntelSwitches.VisionRadius = true
        end
        if bpIntel.RadarRadius and bpIntel.RadarRadius > 0 then
            self.IntelSwitches.RadarRadius = true
        end
        if bpIntel.SonarRadius and bpIntel.SonarRadius > 0 then
            self.IntelSwitches.SonarRadius = true
        end
    end,

    OnIntelEnabled = function(self)
        OldUnit.OnIntelEnabled(self)
        local bp = self:GetBlueprint()
        local cloaked = self:IsIntelEnabled('Cloak') 
        #LOG('*DEBUG: IsIntelEnalbed(Cloak): ', repr(cloaked))
        if cloaked and not self.CloakEffectEnabled then
            #LOG('*DEBUG: Cloak Enabled')
            self.CloakEffectEnabled = true
            self:SetMesh(bp.Display.CloakMeshBlueprint , true)
        end
    end,

    OnIntelDisabled = function(self)
      OldUnit.OnIntelDisabled(self)
        if not self:IsIntelEnabled('Cloak') and self.CloakEffectEnabled then
            self.CloakEffectEnabled = false
            #LOG('*DEBUG: Cloak Disabled')
        self:SetMesh(self:GetBlueprint().Display.MeshBlueprint, true)
        end
    end,


    IntelWatchThread = function(self)
        local energy = 0
        local mass = 0
        local fraction = 1
        local Enabled = true
        while not self:IsDead() do
            WaitSeconds(0.5)
            energy = self:GetConsumptionPerSecondEnergy()
            mass = self:GetConsumptionPerSecondMass()
            if energy == 0 and mass == 0 then
                fraction = 1
            else
                fraction = self:GetResourceConsumed()
                --LOG("|| fraction = " .. fraction)
            end
            if fraction == 1 then
                --LOG("|| EnableUnitIntel")
                if Enabled == false then
                    self:EnableUnitIntel()      -- Make sure this doesn't turn on intel options that have been manually turned off.
                    Enabled = true
                end
            else
                --LOG("|| DisableUnitIntel")
                if Enabled == true then
                    self:DisableUnitIntel()
                    Enabled = false
                end
            end
        end
    end,

    CreateVisionEntity = function(self,army)
        if not self:IsDead() then
            -- Create the entity
            if not self.MyVisionEntities then
                self.MyVisionEntities = {}
            end
            if not self.MyVisionEntities[army] then
                local NewVisionEntity = Entity {}
                self.MyVisionEntities[army] = NewVisionEntity
                table.insert(self.MyVisionEntities,NewVisionEntity)
                NewVisionEntity:InitIntel(army,'Vision',5)
                NewVisionEntity:EnableIntel('Vision')
                NewVisionEntity:AttachBoneTo(-1,self,0)
            end
            -- Kill the entity after 5 seconds in a way that the timer can be reset
            if not self.MyVisionEntityThreads then
                self.MyVisionEntityThreads = {}
            end
            if self.MyVisionEntityThreads[army] then
                self.MyVisionEntityThreads[army]:Destroy()
                self.MyVisionEntityThreads[army] = nil
            end
            self.MyVisionEntityThreads[army] = self:ForkThread(self.StartVisionEntityThread,army)
        end
    end,

    StartVisionEntityThread = function(self,army)
        WaitSeconds(5)
        self.MyVisionEntities[army]:Destroy()
        self.MyVisionEntities[army] = nil
    end,


    ##########################################################################################
    ##  ENERGY MANAGEMENT
    ##########################################################################################

    SetupEnergyUsage = function(self)
        self.EnergySwitch = {}

        self:SetScriptBit('RULEUTC_CloakToggle', true)
        self:SetScriptBit('RULEUTC_JammingToggle', true)
        self:SetScriptBit('RULEUTC_StealthToggle', true)
        self:SetScriptBit('RULEUTC_SpecialToggle', true)
        self:SetScriptBit('RULEUTC_GenericToggle', true)
        self:SetScriptBit('RULEUTC_IntelToggle', true)
        self:SetScriptBit('RULEUTC_ProductionToggle', true)
        self:SetScriptBit('RULEUTC_ShieldToggle', true)
        self.EnergySwitch.Cloak = false
        self.EnergySwitch.CounterIntel = false
        self.EnergySwitch.Custom1 = false
        self.EnergySwitch.Custom2 = false
        self.EnergySwitch.Generic = false
        self.EnergySwitch.Intel = false
        self.EnergySwitch.Production = false
        self.EnergySwitch.Shield = false
        self:UpdateEnergyUsage()

        -------------------------------------------------------------------------------------
--[[
        -- LOGS to track what things are set to
        if self.EnergySwitch.Cloak == true then
            LOG("|| Cloak On")
        else
            LOG("|| Cloak Off")
        end
        if self.EnergySwitch.CounterIntel == true then
            LOG("|| CounterIntel On")
        else
            LOG("|| CounterIntel Off")
        end
        if self.EnergySwitch.Custom1 == true then
            LOG("|| Custom1 On")
        else
            LOG("|| Custom1 Off")
        end
        if self.EnergySwitch.Custom2 == true then
            LOG("|| Custom2 On")
        else
            LOG("|| Custom2 Off")
        end
        if self.EnergySwitch.Generic == true then
            LOG("|| Generic On")
        else
            LOG("|| Generic Off")
        end
        if self.EnergySwitch.Intel == true then
            LOG("|| Intel On")
        else
            LOG("|| Intel Off")
        end
        if self.EnergySwitch.Production == true then
            LOG("|| Production On")
        else
            LOG("|| Production Off")
        end
        if self.EnergySwitch.Shield == true then
            LOG("|| Shield On")
        else
            LOG("|| Shield Off")
        end
--]]
        -------------------------------------------------------------------------------------
    end,

    UpdateEnergyUsage = function(self)
        local Sum = 0
        --LOG("|| Starting at 0")
        if self.EnergySwitch.Cloak == true then
            if self.Stopped == true then
                Sum = Sum + (self.EnergyUsageRates.Cloak or 0)
            else
                if self.EnergyUsageRates.CloakWhileMoving then
                    Sum = Sum + self.EnergyUsageRates.CloakWhileMoving
                else
                    Sum = Sum + (self.EnergyUsageRates.Cloak or 0)
                end
            end
            --LOG("|| * Adding " .. (self.EnergyUsageRates.Cloak or 0) .. " energy usage for Cloak")
        end
        if self.EnergySwitch.CounterIntel == true then
            Sum = Sum + (self.EnergyUsageRates.CounterIntel or 0)
            --LOG("|| * Adding " .. (self.EnergyUsageRates.CounterIntel or 0) .. " energy usage for CounterIntel")
        end
        if self.EnergySwitch.Custom1 == true then
            Sum = Sum + (self.EnergyUsageRates.Custom1 or 0)
            --LOG("|| * Adding " .. (self.EnergyUsageRates.Custom1 or 0) .. " energy usage for Custom1")
        end
        if self.EnergySwitch.Custom2 == true then
            Sum = Sum + (self.EnergyUsageRates.Custom2 or 0)
            --LOG("|| * Adding " .. (self.EnergyUsageRates.Custom2 or 0) .. " energy usage for Custom2")
        end
        if self.EnergySwitch.Generic == true then
            Sum = Sum + (self.EnergyUsageRates.Generic or 0)
            --LOG("|| * Adding " .. (self.EnergyUsageRates.Generic or 0) .. " energy usage for Generic")
        end
        if self.EnergySwitch.Intel == true then
            Sum = Sum + (self.EnergyUsageRates.Intel or 0)
            --LOG("|| * Adding " .. (self.EnergyUsageRates.Intel or 0) .. " energy usage for Intel")
        end
        if self.EnergySwitch.Production == true then
            Sum = Sum + (self.EnergyUsageRates.Production or 0)
            --LOG("|| * Adding " .. (self.EnergyUsageRates.Production or 0) .. " energy usage for Production")
        end
        if self.EnergySwitch.Shield == true then
            Sum = Sum + (self.EnergyUsageRates.Shield or 0)
            --LOG("|| * Adding " .. (self.EnergyUsageRates.Shield or 0) .. " energy usage for Shield")
        end
        self.EnergyMaintenanceConsumptionOverride = Sum
        --LOG("|| Updating energy usage to " .. Sum)
        self:SetMaintenanceConsumptionActive()
        self:UpdateConsumptionValues()
    end,

    OnScriptBitSet = function(self, bit)
        if self.EnergySwitch then
            --LOG("|| Setting bit " .. bit)
            if bit == 0 then # shield toggle
                self.EnergySwitch.Shield = true
                self:EnableShield()
            elseif bit == 1 then # weapon toggle
                # do something
            elseif bit == 2 then # counterintel toggle
                -- Stealth, Jam & Spoof can go together because no unit would have more than one.
                --LOG("|| Setting EnergySwitch.CounterIntel to false")
                self.EnergySwitch.CounterIntel = false
                --LOG("|| Disable RadarStealth")
                self:DisableUnitIntel('RadarStealth')
                --LOG("|| Disable RadarStealthField")
                self:DisableUnitIntel('RadarStealthField')
                --LOG("|| Disable SonarStealth")
                self:DisableUnitIntel('SonarStealth')
                --LOG("|| Disable SonarStealthField")
                self:DisableUnitIntel('SonarStealthField')
                --LOG("|| Disable Jammer")
                self:DisableUnitIntel('Jammer')
                --LOG("|| Disable Spoof")
                self:DisableUnitIntel('Spoof')
            elseif bit == 3 then # intel toggle
                self.EnergySwitch.Intel = false
                --LOG("|| Disable Sonar")
                self:DisableUnitIntel('Sonar')
                --LOG("|| Disable Radar")
                self:DisableUnitIntel('Radar')
                --LOG("|| Disable Omni")
                self:DisableUnitIntel('Omni')
            elseif bit == 4 then # production toggle
                self.EnergySwitch.Production = false
                self:OnProductionPaused()
            elseif bit == 5 then # stealth toggle
                self.EnergySwitch.Custom1 = false
                # FREE TOGGLE!!
            elseif bit == 6 then # generic pause toggle
                self:SetPaused(true)
            elseif bit == 7 then # special toggle
                self.EnergySwitch.Custom2 = true
                self:EnableSpecialToggle()
            elseif bit == 8 then # cloak toggle
                self.EnergySwitch.Cloak = false
                --LOG("|| Disable Cloak")
                self:DisableUnitIntel('Cloak')
                --LOG("|| Disable CloakField")
                self:DisableUnitIntel('CloakField')
            end
            local AllStopped = true
            for k,v in self.EnergySwitch do 
                if v == true then
                    AllStopped = false
                    break
                end
            end
            if AllStopped == true then
                self:StopUnitAmbientSound( 'ActiveLoop' )
            end
            self:UpdateEnergyUsage()
        else
            OldUnit.OnScriptBitSet(self,bit)
        end
    end,

    OnScriptBitClear = function(self, bit)
        if self.EnergySwitch then
            --LOG("|| Clearing bit " .. bit)
            if bit == 0 then # shield toggle
                self.EnergySwitch.Shield = false
                self:DisableShield()
            elseif bit == 1 then # weapon toggle
                # do something
            elseif bit == 2 then # counterintel toggle
                -- Stealth, Jam & Spoof can go together because no unit would have more than one.
                --LOG("|| Setting EnergySwitch.CounterIntel to true")
                self.EnergySwitch.CounterIntel = true
                --LOG("|| Enable RadarStealth")
                self:EnableUnitIntel('RadarStealth')
                --LOG("|| Enable RadarStealthField")
                self:EnableUnitIntel('RadarStealthField')
                --LOG("|| Enable SonarStealth")
                self:EnableUnitIntel('SonarStealth')
                --LOG("|| Enable SonarStealthField")
                self:EnableUnitIntel('SonarStealthField')
                --LOG("|| Enable Jammer")
                self:EnableUnitIntel('Jammer')
                --LOG("|| Enable Spoof")
                self:EnableUnitIntel('Spoof')
            elseif bit == 3 then # intel toggle
                self.EnergySwitch.Intel = true
                --LOG("|| Enable Sonar")
                self:EnableUnitIntel('Sonar')
                --LOG("|| Enable Radar")
                self:EnableUnitIntel('Radar')
                --LOG("|| Enable Omni")
                self:EnableUnitIntel('Omni')
            elseif bit == 4 then # production toggle
                self.EnergySwitch.Production = true
                self:OnProductionUnpaused()
            elseif bit == 5 then # stealth toggle
                self.EnergySwitch.Custom1 = true
                # FREE TOGGLE!!
            elseif bit == 6 then # generic pause toggle
                self:SetPaused(false)
            elseif bit == 7 then # special toggle
                self.EnergySwitch.Custom2 = false
                self:DisableSpecialToggle()
            elseif bit == 8 then # cloak toggle
                self.EnergySwitch.Cloak = true
                --LOG("|| Enable Cloak")
                self:EnableUnitIntel('Cloak')
                --LOG("|| Enable CloakField")
                self:EnableUnitIntel('CloakField')
            end
            local AllStopped = true
            for k,v in self.EnergySwitch do 
                if v == true then
                    AllStopped = false
                    break
                end
            end
            if AllStopped == false then
                self:PlayUnitAmbientSound( 'ActiveLoop' )
            end
            self:UpdateEnergyUsage()
        else
            OldUnit.OnScriptBitClear(self,bit)
        end
    end,

    DisableUnitIntel = function(self, intel)
        local intDisabled = false
        if not self.IntelDisables then return end
        if intel then
            self.IntelDisables[intel] = self.IntelDisables[intel] + 1
            --if self.IntelDisables[intel] == 1 then
                #LOG('*DEBUG: Disabling Intel: ', repr(intel))
                self:DisableIntel(intel)
                intDisabled = true
            --end
        else
            for k, v in self.IntelDisables do
                self.IntelDisables[k] = v + 1
                --if self.IntelDisables[k] == 1 then
                    if k == 'Radar' or k == 'Sonar' or k == 'RadarStealth' or k == 'RadarStealthField' or k == 'SonarStealth' or k == 'SonarStealthField' then
                        -- LOG('|| Disabling Intel: ', repr(k))
                        self:DisableIntel(k)
                        intDisabled = true
                    end
                --end
            end
        end       
        if intDisabled then
            self:OnIntelDisabled()
        end
    end,

    EnableUnitIntel = function(self, intel)
        local layer = self:GetCurrentLayer()
        local bp = self:GetBlueprint()
        local intEnabled = false
        if layer == 'Seabed' or layer == 'Sub' or layer == 'Water' then
            self:EnableIntel('WaterVision')
        end
        if intel then
            --if self.IntelDisables[intel] == 1 then
                self:EnableIntel(intel)
                #LOG('*DEBUG: Enabling Intel: ', repr(intel))
                intEnabled = true
            --end
            self.IntelDisables[intel] = self.IntelDisables[intel] - 1
        else
            for k, v in self.IntelDisables do
                --if v == 1 then
                    if k == 'Radar' or k == 'Sonar' or k == 'RadarStealth' or k == 'RadarStealthField' or k == 'SonarStealth' or k == 'SonarStealthField' then
                        self:EnableIntel(k)
                        --LOG('|| Enabling Intel: ', repr(k))
                        if self:IsIntelEnabled(k) then
                            intEnabled = true
                        end
                    end
                --end
                self.IntelDisables[k] = v - 1
            end
        end

        if not self.IntelThread then
            self.IntelThread = self:ForkThread(self.IntelWatchThread)
        end  
      
        if intEnabled then
            self:OnIntelEnabled()
        end
    end,

    ##########################################################################################
    ##  UTILITIES
    ##########################################################################################

    EnemyInSphere = function(self,radius) 
        pos = self:GetPosition() 
        local units = GetUnitsInRect(self:MakeAreaRectangle(pos,radius)) 
        if not units then 
            return false 
        end 
        local army = self:GetArmy() 
        for k,v in units do 
            if(army != v:GetArmy()) then 
                if not EntityCategoryContains(categories.BENIGN, v) then  -- This should take out decloak due to enemy wall sections.
                    local dist = VDist3(pos,v:GetPosition()) 
                    if dist <= radius then 
                        return true 
                    end 
                end 
            end 
        end 
    end, 

    MakeAreaRectangle = function(self,Position,Radius)
        local x1 = Position.x - Radius 
        local y1 = Position.y - Radius 
        local z1 = Position.z - Radius 
        local x2 = Position.x + Radius 
        local y2 = Position.y + Radius 
        local z2 = Position.z + Radius 
        return Rect(x1, z1, x2, z2)
    end, 
   
    GetNearestEnemyInSphere = function(self,radius) -- This function not used in mod yet.
        local pos = self:GetPosition() 
        local units = GetUnitsInRect(self:MakeAreaRectangle(pos,radius)) 
        local NearestDistance
        local NearestEnemy = {}
        if not units then 
            return {}
        end 
        local army = self:GetArmy() 
        for k,v in units do 
            if(army != v:GetArmy()) then 
                local dist = VDist3(pos,v:GetPosition()) 
                if NearestDistance then
                    if dist < NearestDistance then
                        NearestEnemy = v
                        NearestDistance = dist
                    end
                else
                    NearestEnemy = v
                    NearestDistance = dist
                end 
            end 
        end
        if NearestDistance <= radius then
            return NearestEnemy
        else
            return {}
        end
    end,

    GetEffectsScale = function(self)
        local bp = self:GetBlueprint()
        local Scale = 1
        if bp.Display.EffectScale then
            Scale = math.max(bp.SizeX,bp.SizeY,bp.SizeZ) * bp.Display.EffectScale
        else
            if bp.BlueprintId == "uas0401" then
                Scale = math.max(bp.SizeX,bp.SizeY,bp.SizeZ) * 0.65
            elseif EntityCategoryContains(categories.NAVAL, self) then
                Scale = math.max(bp.SizeX,bp.SizeY,bp.SizeZ) * 0.25
            else
                Scale = math.max(bp.SizeX,bp.SizeY,bp.SizeZ) * 0.75
            end
        end
        return Scale
    end,

    Get2DDistance = function(self,Pos1,Pos2)
        return VDist2(Pos1[1], Pos1[3], Pos2[1], Pos2[3])
    end,

    GetACU = function(self)
        local pos = self:GetPosition() 
        local aiBrain = self:GetAIBrain()   
        self.MyACUs = aiBrain:GetUnitsAroundPoint( categories.COMMAND, pos, 100000, 'Ally' )
    end,

    ##########################################################################################
    ## Barrel Rolls
    ##########################################################################################
    # Using this function will cause a unit to do a single barrel roll.

    StartBarrelRoll = function(self)
        self.StartBarrelRollThread = self:ForkThread( self.BarrelRollThread )
    end,

    BarrelRollThread = function(self)
        #LOG("|| Starting Roll Thread")
        local bp = self:GetBlueprint().Display
        if not self.RollManip and not self:IsDead() then
        #INFO: manip = CreateRotator(unit, bone, axis, [goal], [speed], [accel], [goalspeed])
            self.RollManip = CreateRotator( self, 0, 'z', nil, 50, 50, 400)
            self.Trash:Add(self.RockManip)
            self.RollManip:SetPrecedence(9999999)
            WaitFor( self.RollManip )
            #if self:IsDead() then return end # abort if the unit died
            #self.RollManip:SetTargetSpeed( -(100) ) #Roll the other way
            #WaitFor( self.RollManip )
            if self:IsDead() then return end # abort if the unit died
            self.RollManip:SetTargetSpeed( 400 )
            #Go back to middle
            self.RollManip:SetGoal( 0 )
            WaitFor( self.RollManip )
            if self.RollManip then
              self.RollManip:Destroy()
              self.RollManip = nil
            end
        #LOG("|| Ending Roll thread")
        end
    end,
} TypeClass = Unit
end

do
local OldUnit = Unit
#local UnitText = import('../../lua/ui/game/unittext.lua')
Unit = Class(OldUnit) {
    ##########################################################################################
    ##  VETERANCY
    ##########################################################################################

    OnCreate = function(self)
        OldUnit.OnCreate(self)
        local bp = self:GetBlueprint()
        self.VeteranLevel = 0
        self.xp = 0
        self.XPnextLevel = 15    -- 5 seconds of combat to start leveling up
        self.DPS = bp.dps       -- stores value of blueprint dps on unit
        self.LevelProgress = 0
        self.Sync.LevelProgress = self.LevelProgress
        if bp.Economy.BuildRate then
            self.Sync.BuildRate = bp.Economy.BuildRate
        end
        self.Sync.RegenRate = bp.Defense.RegenRate
    end, 
   --[[ Getting XP for building things - disabling while not used.
    OnStopBeingBuilt = function(self, builder, layer)
        if self:GetBlueprint().Economy.xpTimeStep then
            self:ForkThread(self.XPOverTime)
        end
        OldUnit.OnStopBeingBuilt(self, builder, layer)
    end,
    
    XPOverTime = function(self)
        local waittime = self:GetBlueprint().Economy.xpTimeStep / 100
        WaitSeconds(waittime)
        while not self:IsDead() do 
            self:AddXP(self.XPnextLevel/100)    
            WaitSeconds(waittime)
        end
    end,
    
    startBuildXPThread = function(self)
        local levelPerSecond = self:GetBlueprint().Economy.BuildXPLevelpSecond
        if not levelPerSecond then return end
        WaitSeconds(1)
        while not self:IsDead() do 
            self:AddXP(self.XPnextLevel*levelPerSecond) 
            WaitSeconds(1)
        end
    end,
     
    OnStartBuild = function(self, unitBeingBuilt, order)
        #LOG('unit:OnStartBuild')
        self.BuildXPThread = ForkThread(self.startBuildXPThread, self)
        OldUnit.OnStartBuild(self, unitBeingBuilt, order)    
    end,

    OnStopBuild = function(self, unitBeingBuilt)
        #LOG('unit:OnStopBuild')
        KillThread(self.BuildXPThread)
        OldUnit.OnStopBuild(self, unitBeingBuilt)
    end,

    OnFailedToBuild = function(self)
        #LOG('unit:OnFailedToBuild')
        KillThread(self.BuildXPThread)
        OldUnit.OnFailedToBuild(self)
    end,
    --]]
    #This function should be used for kills made through the script, since kills through the engine (projectiles etc...) are already counted.
    AddKills = function(self, numKills)
        #Add the kills, then check veterancy junk.
        local unitKills = self:GetStat('KILLS', 0).Value + numKills
        self:SetStat('KILLS', unitKills)
        
        --[[ Getting rid of this since veterancy is no longer calculated using kills.
        local vet = self:GetBlueprint().Veteran or Game.VeteranDefault
        
        local vetLevels = table.getsize(vet)
        if self.VeteranLevel == vetLevels then
            return
        end

        local nextLvl = self.VeteranLevel + 1
        local nextKills = vet[('Level' .. nextLvl)]
        
        #Since we could potentially be gaining a lot of kills here, check if we gained more than one level
        while unitKills >= nextKills and self.VeteranLevel ~= vetLevels do
            self:SetVeteranLevel(nextLvl)
            
            nextLvl = self.VeteranLevel + 1
            nextKills = vet[('Level' .. nextLvl)]
        end 
        --]]
    end,

    
    
    # Overriding this to prevent veterancy via kills.
    # Uses SetVeteranLevel instead.
    SetVeterancy = function(self, veteranLevel)
       --[[ veteranLevel = veteranLevel or 0
        if veteranLevel == 0 then
            return
        end
        local bp = self:GetBlueprint()
        if bp.Veteran['Level1'] then
            ##LOG('unit:AddKills')
            self:AddKills(bp.Veteran['Level1']*veteranLevel)
        elseif import('/lua/game.lua').VeteranDefault['Level'..veteranLevel] then
            ##LOG('VeteranDefault.unit:AddKills')
            self:AddKills(import('/lua/game.lua').VeteranDefault['Level'..veteranLevel])
        else
            error('Invalid veteran level - ' .. veteranLevel)
        end --]]
        
    end, 

    #Check to see if we should veteran up.
    CheckVeteranLevel = function(self)
        local bp = self:GetBlueprint()
        #LOG('xp:' .. self.xp.. ' level at:' .. self.XPperLevel)
        while self.xp >= self.XPnextLevel do
            self.xp = self.xp - self.XPnextLevel
            self:SetVeteranLevel(self.VeteranLevel + 1)
            self.XPnextLevel = self.XPnextLevel+((1/(self.XPnextLevel+50))*800)
            --[[ 15
            26.72634403
            37.46937083
            47.619831
            57.35125999
            66.76128635
            75.91217867
            84.8469324
            93.59691667
            102.185967 ]]--

        end
        self.LevelProgress = self.xp / self.XPnextLevel + self.VeteranLevel
        self.Sync.LevelProgress = self.LevelProgress
    end,
    
    #Used by SetVeteranLevel to calculate how much xp each level is. 
    AddLevels = function(self, levels)
        local bp = self:GetBlueprint()
        local curlevel = self.VeteranLevel
        local percent = self.LevelProgress - curlevel
        
        local xpAdd = 0
        if levels >= (1-percent) then
            xpAdd = self.XPnextLevel * (1-percent)
            levels=levels-(1-percent)
        else
            xpAdd =self.XPnextLevel * levels
            levels=0
        end
            
        while levels > 1 do
            levels=levels-1
            curlevel = curlevel +1
            xpAdd=xpAdd + self.XPnextLevel + 5 * self.VeteranLevel
        end
        #Do any remaining % of a level
        xpAdd=xpAdd + self.XPnextLevel + 5 * self.VeteranLevel * levels
        self:AddXP(xpAdd)
    end,
    
    #Set the veteran level to the level specified
    SetVeteranLevel = function(self, level)
        #LOG(' ')
        #LOG('*DEBUG: '.. self:GetBlueprint().Description .. ' VETERAN UP! LEVEL ', repr(level))
        local old = self.VeteranLevel
        self.VeteranLevel = level
        
        # Apply default veterancy buffs
		-- Pawz: Commenting this out Dec 14 2009 to return veterancy values back to somewhat normal.
        --local buffTypes = { 'Regen', 'DPS', 'Damage','DamageArea','Range','Speed','Vision','OmniRadius','Radar','BuildRate','ResourceProduction','RateOfFire','Shield'}
        --for k,bType in buffTypes do
        --    Buff.ApplyBuff( self, 'Veterancy' .. bType)
        --end
        #Create unique unit health buff - 10% of the original health of the unit is added each level
        local bp = self:GetBlueprint()
        local bName = self:GetUnitId() .. 'VetUpgrades'
        #Add Health Veterancy Buff
        if not Buffs[bName] then
            BuffBlueprint {
                MaxLevel = 400,
                Name = bName,
                DisplayName = 'Upgrades from Veterancy',
                BuffType = 'VETERANCY',
                Stacks = 'ALWAYS',
                Duration = -1,
                Affects = {
                    MaxHealth = {
                        Add = bp.Defense.MaxHealth * 0.1,
                        Mult = 1,
                    },
                },
            }
        end
        Buff.ApplyBuff(self, bName)
        
        # Get any overriding buffs if they exist
        local bp = self:GetBlueprint().Buffs
        #Check for unit buffs
        
        if bp then
            for bLevel,bData in bp do
                if (bLevel == 'Any' or bLevel == 'Level'..level) then
                    for bType,bValues in bData do
                        local buffName = self:CreateUnitBuff(bLevel,bType,bValues)
                        if buffName then
                            Buff.ApplyBuff( self, buffName )
                        end
                    end
                end
            end         
        end

        self:GetAIBrain():OnBrainUnitVeterancyLevel(self, level)
        self:DoUnitCallbacks('OnVeteran')
        #Fire off the new level effect
        self:PlayVeteranFx(level)
    end,
    
    #function to generate the new veterancy buffs
    CreateUnitBuff = function(self, levelName, buffType, buffValues)
        
        # Generate a buff based on the unitId
        local buffName = self:GetUnitId() .. levelName .. buffType
        LOG("|| Creating a special unit buff! " .. buffName)
        LOG("||" .. repr(buffValues))
        local buffMinLevel = nil
        local buffMaxLevel = nil
        if buffValues.MinLevel then buffMinLevel = buffValues.MinLevel end
        if buffValues.MaxLevel then buffMaxLevel = buffValues.MaxLevel end
        
        
        # Create the buff if needed
        if not Buffs[buffName] then
            #LOG(buffName .. ': '..buffMinLevel.. ' - '..buffMaxLevel)
            BuffBlueprint {
                MinLevel = buffMinLevel,
                MaxLevel = buffMaxLevel,
                Name = buffName,
                DisplayName = buffName,
                BuffType = buffType,
                Stacks = buffValues.Stacks,
                #self.BuffTypes[buffType].BuffStacks,
                Duration = buffValues.Duration,
                Affects = buffValues.Affects,
            }
        end
        
        # Return the buffname so the buff can be applied to the unit
        return buffName
    end,
    
    AddXP = function(self,amount)
        if not self.XPnextLevel then return end
        self.xp = self.xp + amount
        self:CheckVeteranLevel()
    end,
    
    #Calculates amount of xp based on the modifiable DPS value.
    AddDamageDone = function(self, amount)
        self:AddXP(amount / self.DPS)
        for k,v in self.MyACUs do
            v:AddXP(amount / v.DPS)
        end
    end,
    
    DoTakeDamage = function(self, instigator, amount, vector, damageType)
        local healthBeforeHit =  self:GetHealth()
        local hisArmy = false
        if instigator and IsUnit(instigator) and instigator:GetArmy() then
            hisArmy = instigator:GetArmy()
        end       
        local myArmy = self:GetArmy()
        
        #Not allowed to get veterancy by killing your own units!
        if hisArmy and myArmy != hisArmy then
            if not EntityCategoryContains(categories.BENIGN, self) then
                if instigator and IsUnit(instigator) and not instigator:IsDead() then
                    local preAdjHealth = self:GetHealth()
                    if amount>=preAdjHealth then
                        #killed the unit, update killer
                        instigator:AddDamageDone(preAdjHealth)
                    else
                        #unit is still alive - both get xp
                        instigator:AddDamageDone(amount)
                        self:AddDamageDone(amount / 4)
                    end
                end
            end
        end
        
        #do the actual damaging bit
        OldUnit.DoTakeDamage(self,instigator,amount,vector,damageType)
    end,
    

    #PlayVeteranFx is staying here for backward compatibility... (original Unit.lua function)   
    PlayVeteranFx = function(self, newLvl)   
        self:CreateVeteranEmitter()   
    end,   
    
    #If you want to hide & show this effect with a toggle, just call   
    #Create and Destroy...   
    CreateVeteranEmitter = function(self)   
        if self.vetEmitter != nil then   
            self.vetEmitter:Destroy()   
            self.vetEmitter = nil   
        end   
        local VetScale = 0
        VetScale = 1+ (self.VeteranLevel*0.2 )
        self:GetEmitterPath()
        self.vetEmitter = CreateAttachedEmitter(self, -1, self:GetArmy(),self:GetEmitterPath() ):ScaleEmitter((self:GetEffectsScale()) * VetScale)   
    end,   
    
    GetEmitterPath = function(self)
        local EmitterPath = '/mods/Supremilated_ATWMedia_20_007/hook/effects/emitters/'
        if EntityCategoryContains(categories.SERAPHIM, self) then
            EmitterPath = EmitterPath .. 'SeraphimVet.bp'
        elseif EntityCategoryContains(categories.UEF, self) then
            EmitterPath = EmitterPath .. 'UEFVet.bp'
        elseif EntityCategoryContains(categories.CYBRAN, self) then
            EmitterPath = EmitterPath .. 'CybranVet.bp'
        elseif EntityCategoryContains(categories.AEON, self) then
            EmitterPath = EmitterPath .. 'AeonVet.bp'
        end
        return EmitterPath
    end,
    
    DestroyVeteranEmitter = function(self)   
        if self.vetEmitter != nil then   
            self.vetEmitter:Destroy()   
            self.vetEmitter = nil   
        end   
    end,   
    
    GetShield = function(self)
        return self.MyShield or nil
    end,

} TypeClass = Unit
end