do
local Buff = import('/lua/sim/Buff.lua')
local OldUnit = URL0101
local CWalkingLandUnit = import('/lua/cybranunits.lua').CWalkingLandUnit
URL0101 = Class(OldUnit) {
    RadarMode = false,
    InMotion = false,
    DoneBeingBuilt = false,
    CurrentRadar = 0,
    OnStopBeingBuilt = function(self, builder, layer)
        CWalkingLandUnit.OnStopBeingBuilt(self,builder,layer)
        CreateRotator(self, 'Spinner', 'y', nil, 90, 5, 90)
        self:SetMaintenanceConsumptionInactive()
        self:SetScriptBit('RULEUTC_CloakToggle', true)
        self:SetScriptBit('RULEUTC_JammingToggle', false)
        self:SetScriptBit('RULEUTC_IntelToggle', false) 
        self:SetScriptBit('RULEUTC_SpecialToggle', false) 
        self:RequestRefreshUI()
        self.RecloakCountdown = 0
        self.CloakManagerThread = self:ForkThread(self.CloakManager)
        self:AddCommandCap('RULEUCC_Move')
        local bpIntel = self:GetBlueprint().Intel
        self.CurrentRadar = bpIntel.RadarRadius
        self:ForkThread(self.WatchEnemyUnits,self)
        self.DoneBeingBuilt = true
    end,
    OnMotionHorzEventChange = function(self, new, old)
        OldUnit.OnMotionHorzEventChange(self, new, old)
        if old == 'Stopped' then
            self:RemoveToggleCap('RULEUTC_SpecialToggle')
            self:RemoveToggleCap('RULEUTC_CloakToggle')
        elseif new == 'Stopped' then
            self:AddToggleCap('RULEUTC_CloakToggle')
            self:AddToggleCap('RULEUTC_SpecialToggle')
        end
    end,
    EnableSpecialToggle = function(self)
        OldUnit.EnableSpecialToggle(self)
        if self.DoneBeingBuilt == true then
            if self.UndeployThread then
                self.UndeployThread:Destroy()
            end
            self:RemoveCommandCap('RULEUCC_Move')
            self:SetImmobile(true)
            self.DeployThread = self:ForkThread(self.Deploy)
        end
    end,
    DisableSpecialToggle = function(self)
        OldUnit.DisableSpecialToggle(self)
        if self.DoneBeingBuilt == true then
            if self.DeployThread then
                self.DeployThread:Destroy()
            end
            self.UndeployThread = self:ForkThread(self.Undeploy)
        end
    end,
    Undeploy = function(self)
        local bpIntel = self:GetBlueprint().Intel
        local MaxRadar = self.CurrentRadar
        local MinRadar = bpIntel.RadarRadius
        local StepSize = 0.25
        local RadarStepSize = (bpIntel.RadarRadiusBoost - bpIntel.RadarRadius) / 20
        while self.CurrentRadar > MinRadar do
            self.CurrentRadar = self.CurrentRadar - RadarStepSize
            BuffBlueprint {
                    Name = 'RadarUnDeploy',
                    DisplayName = 'Adds range when scout is deployed',
                    BuffType = 'RadarUnDeploy',
                    Stacks = 'ALWAYS',
                    Duration = -1,
                    Affects = {
                        RadarRadius = {
                            Add = (0 - RadarStepSize),
                            Mult = 1,
                        },
                    },
                }
            Buff.ApplyBuff(self,'RadarUnDeploy')

            WaitSeconds(StepSize)
        end
        Buff.RemoveBuff(self,'RadarDeploy')
        Buff.RemoveBuff(self,'RadarUnDeploy')
        self:AddCommandCap('RULEUCC_Move')
            self:SetImmobile(false)
    end,
    Deploy = function(self)
        local bpIntel = self:GetBlueprint().Intel
        local MaxRadar = bpIntel.RadarRadiusBoost
        local MinRadar = self.CurrentRadar
        local StepSize = 0.25
        local RadarStepSize = (bpIntel.RadarRadiusBoost - bpIntel.RadarRadius) / 20
        while self.CurrentRadar < MaxRadar do
            self.CurrentRadar = self.CurrentRadar + RadarStepSize
            BuffBlueprint {
                    Name = 'RadarDeploy',
                    DisplayName = 'Adds range when scout is deployed',
                    BuffType = 'RadarDeploy',
                    Stacks = 'ALWAYS',
                    Duration = -1,
                    Affects = {
                        RadarRadius = {
                            Add = RadarStepSize,
                            Mult = 1,
                        },
                    },
                }
            Buff.ApplyBuff(self,'RadarDeploy')

            WaitSeconds(StepSize)
        end
    end,
    WatchEnemyUnits = function(self)
        while not self:IsDead() do
            local pos = self:GetPosition() 
            local aiBrain = self:GetAIBrain()   
            local units = aiBrain:GetUnitsAroundPoint(categories.ALLUNITS, pos, 30, 'Enemy' )
            for k,v in units do
                if v:GetBlueprint().Intel then
                    bpIntel = v:GetBlueprint().Intel
                    if bpIntel.RadarRadius then
                        BuffBlueprint {
                                Name = 'RadarNerf',
                                DisplayName = 'Cuts Radar to 25%',
                                BuffType = 'RadarNerf',
                                Stacks = 'REPLACE',
                                Duration = 5,
                                Affects = {
                                    RadarRadius = {
                                        Add = 0,
                                        Mult = 0.25,
                                    },
                                },
                            }
                        Buff.ApplyBuff(v,'RadarNerf')
                    end
                end
            end
            WaitSeconds(4)
        end
    end,
}   
TypeClass = URL0101
end
