do
local EffectTemplate = import('/lua/EffectTemplates.lua')
local OldUnit = XRL0305
XRL0305 = Class(OldUnit) {
    InMotion = false,
    DoneBeingBuilt = false,
    OnStopBeingBuilt = function(self, builder, layer)
        OldUnit.OnStopBeingBuilt(self, builder, layer)
        self:SetScriptBit('RULEUTC_SpecialToggle', false) 
        self:AddCommandCap('RULEUCC_Move')
        self.DoneBeingBuilt = true
    end,
    OnMotionHorzEventChange = function(self, new, old)
        OldUnit.OnMotionHorzEventChange(self, new, old)
        if old == 'Stopped' then
            self:RemoveToggleCap('RULEUTC_SpecialToggle')
        elseif new == 'Stopped' then
            self:AddToggleCap('RULEUTC_SpecialToggle')
        end
    end,
    EnableSpecialToggle = function(self)
        OldUnit.EnableSpecialToggle(self)
        if self.DoneBeingBuilt == true then
            if self.UndeployThread then
                self.UndeployThread:Destroy()
            end
            self:RemoveCommandCap('RULEUCC_Move')
            self:SetImmobile(true)
            self.DeployThread = self:ForkThread(self.Deploy)
        end
    end,
    DisableSpecialToggle = function(self)
        OldUnit.DisableSpecialToggle(self)
        if self.DoneBeingBuilt == true then
            if self.DeployThread then
                self.DeployThread:Destroy()
            end
            self.UndeployThread = self:ForkThread(self.Undeploy)
        end
    end,
    Undeploy = function(self)
        self:CreateTransitionEffects()
        self:RemoveToggleCap('RULEUTC_SpecialToggle')
        self:SetWeaponEnabledByLabel('Disintigrator', false)
        WaitSeconds(3)
        self:SetWeaponEnabledByLabel('Disintigrator', true)
        self:CleanupTransitionEffects()
        self:AddToggleCap('RULEUTC_SpecialToggle')
        self:AddCommandCap('RULEUCC_Move')
        self:SetImmobile(false)
        self:SetUnarmoredState()
        self:CleanupDeployedEffects()
    end,
    Deploy = function(self)
        self:CreateTransitionEffects()
        self:RemoveToggleCap('RULEUTC_SpecialToggle')
        self:SetWeaponEnabledByLabel('Disintigrator', false)
        WaitSeconds(20)
        self:SetWeaponEnabledByLabel('Disintigrator', true)
        self:CleanupTransitionEffects()
        self:CreateDeployedEffects()
        self:AddToggleCap('RULEUTC_SpecialToggle')
        self:SetArmoredState()
    end,
    CreateDeployedEffects = function(self)
        local army = self:GetArmy()
        self.DeployEffectBag = {}
        local Scale = self:GetEffectsScale()
        for k, v in EffectTemplate.BrickDeployment do
            local fx = CreateAttachedEmitter(self, -1, army,v):ScaleEmitter(Scale)
            self.Trash:Add(fx)
            table.insert( self.DeployEffectBag, fx)
        end
    end,
    CleanupDeployedEffects = function( self )
        if self.DeployEffectBag then
            for keys,values in self.DeployEffectBag do
                values:Destroy()
            end
            self.DeployEffectBag = {}   -- or nil?
        end
    end,
    CreateTransitionEffects = function(self)
        local army = self:GetArmy()
        self.TransitionEffectBag = {}
        local Scale = self:GetEffectsScale()
        for k, v in EffectTemplate.BrickTransition do
            local fx = CreateAttachedEmitter(self, -1, army,v):ScaleEmitter(Scale)
            self.Trash:Add(fx)
            table.insert( self.TransitionEffectBag , fx)
        end
    end,
    CleanupTransitionEffects = function( self )
        if self.TransitionEffectBag then
            for keys,values in self.TransitionEffectBag do
                values:Destroy()
            end
            self.TransitionEffectBag = {}   -- or nil?
        end
    end,
}   
TypeClass = XRL0305
end
