do
local EffectTemplate = import('/lua/EffectTemplates.lua')
local DefaultUnitsFile = import('/lua/defaultunits.lua')
local OldUnit = XSL0001 
XSL0001 = Class(OldUnit) { 

    CurrentWeaponRange = 0,
    MaxNosTotal = 0,
    MaxNosUnit = 0,

    OnCreate = function(self)
        OldUnit.OnCreate(self)
        self:AddBuildRestriction( categories.SERAPHIM * ( categories.BUILTBYTIER1ENGINEER + categories.BUILTBYTIER2COMMANDER + categories.BUILTBYTIER3COMMANDER + categories.EXPERIMENTAL) )
        local cat = ParseEntityCategory('BUILTBYCOMMANDER SERAPHIM')
        self:RemoveBuildRestriction(cat)
    end,

    OnStopBeingBuilt = function(self, builder, layer)
        OldUnit.OnStopBeingBuilt(self,builder,layer)
        self.CurrentWeaponRange = self:GetBlueprint().Weapon[1].MaxRadius
    end,
    
    OnStartBuild = function(self, unitBeingBuilt, order)
        DefaultUnitsFile.MobileUnit.OnStartBuild(self,unitBeingBuilt, order)
        OldUnit.OnStartBuild(self, unitBeingBuilt, order)
    end,
    
    NosWatch = function(self)
        local NosHitsPerSecond = 5
        while not self:IsDead() do
            if self:GetHealth() < self:GetMaxHealth() then
                local LiveVictims = self:GetVictims()
                local LiveVictimCount = self:CountObjectsInTable(LiveVictims)
                local HealthNeeded = self:GetMaxHealth() - self:GetHealth()
                while HealthNeeded > 0 and LiveVictimCount > 0 do
                    local NosPerUnit = math.min(self.MaxNosUnit,math.min(HealthNeeded,self.MaxNosTotal) / LiveVictimCount) / NosHitsPerSecond
                    local TotalTransfer = 0
                    for k,v in LiveVictims do
                        local DamageAmount = math.ceil(math.min(NosPerUnit,v:GetHealth()))
                        v:DoTakeDamage(self,DamageAmount,VDiff(self:GetPosition(), v:GetPosition()),'Normal')
                        TotalTransfer = TotalTransfer + DamageAmount
                        if not v.NOSEffect then
                            self:CreateNOSEffects(v)
                            v.NOSEffect = true
                        end
                        if v.NOSTimeoutThread then
                            v.NOSTimeoutThread:Destroy()
                            v.NOSTimeoutThread = nil
                        end
                        v.NOSTimeoutThread = v:ForkThread(v.NOSTimeout, v)
                    end
                    self:AdjustHealth(self, TotalTransfer)
                    WaitSeconds(0.2)
                    LiveVictims = self:GetVictims()
                    LiveVictimCount = self:CountObjectsInTable(LiveVictims)
                    HealthNeeded = self:GetMaxHealth() - self:GetHealth()
                end
            end
            WaitSeconds(2)
        end
    end,

    CountObjectsInTable = function(self,TableObject)
        local Count = 0
        for k,v in TableObject do
            Count = Count + 1
        end
        return Count
    end,

    GetVictims = function(self)
        local position = self:GetPosition()
        local radius = self.CurrentWeaponRange
        local x1 = position.x - radius
        local y1 = position.y - radius
        local z1 = position.z - radius
        local x2 = position.x + radius
        local y2 = position.y + radius
        local z2 = position.z + radius
        local UnitsinRec = GetUnitsInRect( Rect(x1, z1, x2, z2) )
        if not UnitsinRec then
            return UnitsinRec
        end
        local RadEntities = {}
        for k, v in UnitsinRec do
            if v != self then
                local dist = VDist3(position, v:GetPosition())
                if dist <= radius then
                    table.insert(RadEntities, v)
                end
            end
        end
        return RadEntities
    end,

    CreateNOSEffects = function(self,TargetUnit)
        local army = TargetUnit:GetArmy()
        TargetUnit.NOSEffectBag = {}
        local Scale = TargetUnit:GetEffectsScale()
        for k, v in EffectTemplate.NOSEffect do
            local fx = CreateAttachedEmitter(TargetUnit, -1, army,v):ScaleEmitter(Scale*.7)
            TargetUnit.Trash:Add(fx)
            table.insert( TargetUnit.NOSEffectBag, fx)
        end
    end,

    TeleportUpgradeDisabled = false,

    DoTakeDamage = function(self, instigator, amount, vector, damageType)
        if self.HasTeleportUpgrade == true then
            if not self:IsDead() then
                if self.DisableTeleportThread then
                    self.DisableTeleportThread:Destroy()
                    self.DisableTeleportThread = nil
                end
                self.DisableTeleportThread = self:ForkThread(self.DisableTeleportUpgrade, self)
            end
        end
        OldUnit.DoTakeDamage(self,instigator,amount,vector,damageType)
    end,

    DisableTeleportUpgrade = function(self)
        self.TeleportUpgradeDisabled = true
        WaitSeconds(3)
        self.TeleportUpgradeDisabled = false
    end, 

    CreateEnhancement = function(self, enh)
        SWalkingLandUnit.CreateEnhancement(self, enh)
        local bp = self:GetBlueprint().Enhancements[enh]

        --                                                         
        --                        BACK                             
        --                                                         

        if enh == 'DamageStabilization' then
            if not Buffs['SeraphimACUDamageStabilization'] then
                BuffBlueprint {
                    Name = 'SeraphimACUDamageStabilization',
                    DisplayName = 'SeraphimACUDamageStabilization',
                    BuffType = 'ACUUPGRADEDMG',
                    Stacks = 'ALWAYS',
                    Duration = -1,
                    Affects = {
                        MaxHealth = {
                            Add = bp.NewHealth,
                            Mult = 1.0,
                        },
                        Regen = {
                            Add = bp.NewRegenRate,
                            Mult = 1.0,
                        },
                    },
                } 
            end
            if Buff.HasBuff( self, 'SeraphimACUDamageStabilization' ) then
                Buff.RemoveBuff( self, 'SeraphimACUDamageStabilization' )
            end  
            Buff.ApplyBuff(self, 'SeraphimACUDamageStabilization')    
        elseif enh == 'DamageStabilizationRemove' then
            if Buff.HasBuff( self, 'SeraphimACUDamageStabilization' ) then
                Buff.RemoveBuff( self, 'SeraphimACUDamageStabilization' )
            end           
        elseif enh == 'DamageStabilizationAdvanced' then
            if not Buffs['SeraphimACUDamageStabilizationAdv'] then
               BuffBlueprint {
                    Name = 'SeraphimACUDamageStabilizationAdv',
                    DisplayName = 'SeraphimACUDamageStabilizationAdv',
                    BuffType = 'ACUUPGRADEDMG',
                    Stacks = 'ALWAYS',
                    Duration = -1,
                    Affects = {
                        MaxHealth = {
                            Add = bp.NewHealth,
                            Mult = 1.0,
                        },
                        Regen = {
                            Add = bp.NewRegenRate,
                            Mult = 1.0,
                        },
                    },
                } 
            end
            if Buff.HasBuff( self, 'SeraphimACUDamageStabilizationAdv' ) then
                Buff.RemoveBuff( self, 'SeraphimACUDamageStabilizationAdv' )
            end  
            Buff.ApplyBuff(self, 'SeraphimACUDamageStabilizationAdv')           
        elseif enh == 'DamageStabilizationAdvancedRemove' then
            # since there's no way to just remove an upgrade anymore, if we're remove adv, we're removing both
            if Buff.HasBuff( self, 'SeraphimACUDamageStabilizationAdv' ) then
                Buff.RemoveBuff( self, 'SeraphimACUDamageStabilizationAdv' )
            end
            if Buff.HasBuff( self, 'SeraphimACUDamageStabilization' ) then
                Buff.RemoveBuff( self, 'SeraphimACUDamageStabilization' )
            end         
        elseif enh == 'Nosferatu' then
            self.MaxNosTotal = bp.MaxNosTotal or 1000
            self.MaxNosUnit = bp.MaxNosUnit or 100
            self.NosThread = self:ForkThread(self.NosWatch)
        elseif enh == 'NosferatuRemove' then
            if self.NosThread then
                self.NosThread:Destroy()
                if Buff.HasBuff( self, 'SeraphimACUDamageStabilizationAdv' ) then
                    Buff.RemoveBuff( self, 'SeraphimACUDamageStabilizationAdv' )
                end
                if Buff.HasBuff( self, 'SeraphimACUDamageStabilization' ) then
                    Buff.RemoveBuff( self, 'SeraphimACUDamageStabilization' )
                end         
            end
        elseif enh == 'Missile' then
            self:AddCommandCap('RULEUCC_Tactical')
            self:AddCommandCap('RULEUCC_SiloBuildTactical')        
            self:SetWeaponEnabledByLabel('Missile', true)
            local wep = self:GetWeaponByLabel('Missile')
            wep:ChangeMaxRadius(256)
            wep:ChangeMinRadius(5)
        elseif enh == 'MissileRemove' then
            self:RemoveCommandCap('RULEUCC_Tactical')
            self:RemoveCommandCap('RULEUCC_SiloBuildTactical')        
            self:SetWeaponEnabledByLabel('Missile', false)
            local wep = self:GetWeaponByLabel('Missile')
            wep:ChangeMaxRadius(0)
            wep:ChangeMinRadius(0)
        elseif enh == 'Teleporter' then
            self.HasTeleportUpgrade = true
            self.InBeaconProximityTesting = false
            self.TeleportRechargeTime = 15
            self.TeleportEnergyCost = 3000
            self:AddCommandCap('RULEUCC_Teleport')
        elseif enh == 'TeleporterRemove' then
            self.HasTeleportUpgrade = false
            self.InBeaconProximityTesting = false
            self:RemoveCommandCap('RULEUCC_Teleport')
            self.TeleportRechargeTime = 60
            self.TeleportEnergyCost = 15000


        --                                                         
        --                      LEFT ARM                           
        --                                                         

        elseif enh =='Nanolathe' then
            local bp = self:GetBlueprint().Enhancements[enh]
            if not bp then return end
            local cat = ParseEntityCategory(bp.BuildableCategoryAdds)
            self:RemoveBuildRestriction(cat)
            if not Buffs['SeraphimACUNanolatheBuildRate'] then
                BuffBlueprint {
                    Name = 'SeraphimACUNanolatheBuildRate',
                    DisplayName = 'SeraphimACUNanolatheBuildRate',
                    BuffType = 'ACUBUILDRATE',
                    Stacks = 'REPLACE',
                    Duration = -1,
                    Affects = {
                        BuildRate = {
                            Add =  bp.NewBuildRate - self:GetBlueprint().Economy.BuildRate,
                            Mult = 1,
                        },
                        MaxHealth = {
                            Add = bp.NewHealth,
                            Mult = 1.0,
                        },
                        Regen = {
                            Add = bp.NewRegenRate,
                            Mult = 1.0,
                        },
                    },
                }
            end
            Buff.ApplyBuff(self, 'SeraphimACUNanolatheBuildRate')
        elseif enh =='NanolatheRemove' then
            local bp = self:GetBlueprint().Economy.BuildRate
            if not bp then return end
            self:RestoreBuildRestrictions()
            self:AddBuildRestriction(categories.SERAPHIM * (categories.BUILTBYTIER2COMMANDER + categories.BUILTBYTIER3COMMANDER + categories.EXPERIMENTAL))
            if Buff.HasBuff( self, 'SeraphimACUNanolatheBuildRate' ) then
                Buff.RemoveBuff( self, 'SeraphimACUNanolatheBuildRate' )
            end
        elseif enh =='AdvancedEngineering' then
            local bp = self:GetBlueprint().Enhancements[enh]
            if not bp then return end
            local cat = ParseEntityCategory(bp.BuildableCategoryAdds)
            self:RemoveBuildRestriction(cat)
            if not Buffs['SeraphimACUT2BuildRate'] then
                BuffBlueprint {
                    Name = 'SeraphimACUT2BuildRate',
                    DisplayName = 'SeraphimACUT2BuildRate',
                    BuffType = 'ACUBUILDRATE',
                    Stacks = 'REPLACE',
                    Duration = -1,
                    Affects = {
                        BuildRate = {
                            Add =  bp.NewBuildRate - self:GetBlueprint().Economy.BuildRate,
                            Mult = 1,
                        },
                        MaxHealth = {
                            Add = bp.NewHealth,
                            Mult = 1.0,
                        },
                        Regen = {
                            Add = bp.NewRegenRate,
                            Mult = 1.0,
                        },
                    },
                }
            end
            Buff.ApplyBuff(self, 'SeraphimACUT2BuildRate')
        elseif enh =='AdvancedEngineeringRemove' then
            local bp = self:GetBlueprint().Economy.BuildRate
            if not bp then return end
            self:RestoreBuildRestrictions()
            self:AddBuildRestriction(categories.SERAPHIM * (categories.BUILTBYTIER2COMMANDER + categories.BUILTBYTIER3COMMANDER + categories.EXPERIMENTAL))
            if Buff.HasBuff( self, 'SeraphimACUT2BuildRate' ) then
                Buff.RemoveBuff( self, 'SeraphimACUT2BuildRate' )
            end
        elseif enh =='T3Engineering' then
            local bp = self:GetBlueprint().Enhancements[enh]
            if not bp then return end
            local cat = ParseEntityCategory(bp.BuildableCategoryAdds)
            self:RemoveBuildRestriction(cat)
            if not Buffs['SeraphimACUT3BuildRate'] then
                BuffBlueprint {
                    Name = 'SeraphimACUT3BuildRate',
                    DisplayName = 'SeraphimCUT3BuildRate',
                    BuffType = 'ACUBUILDRATE',
                    Stacks = 'REPLACE',
                    Duration = -1,
                    Affects = {
                        BuildRate = {
                            Add =  bp.NewBuildRate - self:GetBlueprint().Economy.BuildRate,
                            Mult = 1,
                        },
                        MaxHealth = {
                            Add = bp.NewHealth,
                            Mult = 1.0,
                        },
                        Regen = {
                            Add = bp.NewRegenRate,
                            Mult = 1.0,
                        },
                    },
                }
            end
            Buff.ApplyBuff(self, 'SeraphimACUT3BuildRate')
        elseif enh =='T3EngineeringRemove' then
            local bp = self:GetBlueprint().Economy.BuildRate
            if not bp then return end
            self:RestoreBuildRestrictions()
            if Buff.HasBuff( self, 'SeraphimACUT3BuildRate' ) then
                Buff.RemoveBuff( self, 'SeraphimACUT3BuildRate' )
            end
            self:AddBuildRestriction(categories.SERAPHIM * (categories.BUILTBYTIER2COMMANDER + categories.BUILTBYTIER3COMMANDER + categories.EXPERIMENTAL))
        elseif enh =='T4Engineering' then
            local bp = self:GetBlueprint().Enhancements[enh]
            if not bp then return end
            local cat = ParseEntityCategory(bp.BuildableCategoryAdds)
            self:RemoveBuildRestriction(cat)
            if not Buffs['SeraphimACUT4BuildRate'] then
                BuffBlueprint {
                    Name = 'SeraphimACUT4BuildRate',
                    DisplayName = 'SeraphimCUT4BuildRate',
                    BuffType = 'ACUBUILDRATE',
                    Stacks = 'REPLACE',
                    Duration = -1,
                    Affects = {
                        BuildRate = {
                            Add =  bp.NewBuildRate - self:GetBlueprint().Economy.BuildRate,
                            Mult = 1,
                        },
                        MaxHealth = {
                            Add = bp.NewHealth,
                            Mult = 1.0,
                        },
                        Regen = {
                            Add = bp.NewRegenRate,
                            Mult = 1.0,
                        },
                    },
                }
            end
            Buff.ApplyBuff(self, 'SeraphimACUT4BuildRate')
        elseif enh =='T4EngineeringRemove' then
            local bp = self:GetBlueprint().Economy.BuildRate
            if not bp then return end
            self:RestoreBuildRestrictions()
            if Buff.HasBuff( self, 'SeraphimACUT4BuildRate' ) then
                Buff.RemoveBuff( self, 'SeraphimACUT4BuildRate' )
            end
            self:AddBuildRestriction(categories.SERAPHIM * (categories.BUILTBYTIER2COMMANDER + categories.BUILTBYTIER3COMMANDER + categories.EXPERIMENTAL))
        elseif enh == 'BlastAttack' then
            if not Buffs['BlastAttack'] then
                BuffBlueprint {
                    Name = 'BlastAttack',
                    DisplayName = 'BlastAttack',
                    BuffType = 'BlastAttack',
                    Stacks = 'ALWAYS',
                    Duration = -1,
                    Affects = {
                        DamageRadius = {
                            Add =  bp.NewDamageRadius or 5,
                            Mult = 1,
                            ByName = 'ChronotronCannon',
                        },
                    },
                }
            end
            Buff.ApplyBuff(self, 'SeraphimACUT4BuildRate')
            #local wep = self:GetWeaponByLabel('ChronotronCannon')
            #wep:AddDamageRadiusMod(bp.NewDamageRadius or 5)
        elseif enh == 'BlastAttackRemove' then
            #local wep = self:GetWeaponByLabel('ChronotronCannon')
            #wep:AddDamageRadiusMod(bp.NewDamageRadius or 5)
            if Buff.HasBuff( self, 'BlastAttack' ) then
                Buff.RemoveBuff( self, 'BlastAttack' )
            end
            
        --                                                         
        --                     RIGHT ARM                           
        --                                                         

        elseif enh == 'WeaponEnhanceLevel1' or enh == 'WeaponEnhanceLevel2' or enh == 'WeaponEnhanceLevel3' or enh == 'WeaponEnhanceLevel4'then
            if not Buffs[enh] then
                BuffBlueprint {
                    Name = enh,
                    DisplayName = enh,
                    BuffType = 'WeaponEnhance',
                    Stacks = 'REPLACE',
                    Duration = -1,
                    Affects = {
                        RateOfFireBuf = {
                            Add = bp.NewRateOfFire,
                            Mult = 1,
                        },
                        MaxRadius = {
                            Add =  bp.NewMaxRadius,
                            Mult = 1,                      
                        },
                        Damage = {
                            Add =  bp.AdditionalDamage,
                            Mult = 1,
                        },
                        DPS = {
                            Add = 0,
                            Mult =  bp.DPSMultiplier,
                        },
                    },
                }
            end
            Buff.ApplyBuff(self, enh)
        elseif enh == 'WeaponEnhanceLevel1Remove' or enh == 'WeaponEnhanceLevel2Remove' or enh == 'WeaponEnhanceLevel3Remove' or enh == 'WeaponEnhanceLevel4Remove' then
            if Buff.HasBuff( self, string.gsub(enh, 'Remove', '') ) then
                Buff.RemoveBuff( self, string.gsub(enh, 'Remove', '') )
            end
        elseif enh == 'RegenAura' then
            local bp = self:GetBlueprint().Enhancements[enh]
            if not Buffs['SeraphimACURegenAura'] then
                BuffBlueprint {
                    Name = 'SeraphimACURegenAura',
                    DisplayName = 'SeraphimACURegenAura',
                    BuffType = 'COMMANDERAURA',
                    Stacks = 'REPLACE',
                    Duration = 5,
                    Affects = {
                        RegenPercent = {
                            Add = 0,
                            Mult = bp.RegenPerSecond or 0.1,
                            Ceil = bp.RegenCeiling,
                            Floor = bp.RegenFloor,
                        },
                    },
                }
            end
            table.insert( self.ShieldEffectsBag, CreateAttachedEmitter( self, 'XSL0001', self:GetArmy(), '/effects/emitters/seraphim_regenerative_aura_01_emit.bp' ) )
            self.RegenThreadHandle = self:ForkThread(self.RegenBuffThread)
        elseif enh == 'RegenAuraRemove' then
            if self.ShieldEffectsBag then
                for k, v in self.ShieldEffectsBag do
                    v:Destroy()
                end
                self.ShieldEffectsBag = {}
            end
            KillThread(self.RegenThreadHandle)
        elseif enh == 'AdvancedRegenAura' then
            if self.RegenThreadHandle then
                if self.ShieldEffectsBag then
                    for k, v in self.ShieldEffectsBag do
                        v:Destroy()
                    end
                    self.ShieldEffectsBag = {}
                end
                KillThread(self.RegenThreadHandle)
                
            end
            local bp = self:GetBlueprint().Enhancements[enh]
            if not Buffs['SeraphimAdvancedACURegenAura'] then
                BuffBlueprint {
                    Name = 'SeraphimAdvancedACURegenAura',
                    DisplayName = 'SeraphimAdvancedACURegenAura',
                    BuffType = 'COMMANDERAURA',
                    Stacks = 'REPLACE',
                    Duration = 5,
                    Affects = {
                        RegenPercent = {
                            Add = 0,
                            Mult = bp.RegenPerSecond or 0.1,
                            Ceil = bp.RegenCeiling,
                            Floor = bp.RegenFloor,
                        },
                        MaxHealth = {
                            Add = 0,
                            Mult = bp.MaxHealthFactor or 1.0,
                        },                        
                    },
                }
            end
            table.insert( self.ShieldEffectsBag, CreateAttachedEmitter( self, 'XSL0001', self:GetArmy(), '/effects/emitters/seraphim_regenerative_aura_01_emit.bp' ) )
            self.AdvancedRegenThreadHandle = self:ForkThread(self.AdvancedRegenBuffThread)
        elseif enh == 'AdvancedRegenAuraRemove' then
            if self.ShieldEffectsBag then
                for k, v in self.ShieldEffectsBag do
                    v:Destroy()
                end
                self.ShieldEffectsBag = {}
            end
            KillThread(self.AdvancedRegenThreadHandle)
        end
    end,

} 
TypeClass = XSL0001
end 
