do
local SWalkingLandUnit = import('/lua/seraphimunits.lua').SWalkingLandUnit
local SDFPhasicAutoGunWeapon = import('/lua/seraphimweapons.lua').SDFPhasicAutoGunWeapon
local OldUnit = XSL0101
XSL0101 = Class(SWalkingLandUnit) {
    RadarMode = false,
    InMotion = false,
    DoneBeingBuilt = false,
    CurrentRadar = 0,

    Weapons = {
        LaserTurret = Class(SDFPhasicAutoGunWeapon) {
            OnWeaponFired = function(self, target)
                local bp = self:GetBlueprint()
                if self.unit.RecloakCountdown < (bp.RecloakAfterFiringDelay or 0) then
                    self.unit.RecloakCountdown = bp.RecloakAfterFiringDelay or 0
                end
                SDFPhasicAutoGunWeapon.OnWeaponFired(self, target)
            end,
        },
    },

    OnStopBeingBuilt = function(self, builder, layer)
        SWalkingLandUnit.OnStopBeingBuilt(self, builder, layer)
        self:SetScriptBit('RULEUTC_IntelToggle', false) 
        self:SetScriptBit('RULEUTC_SpecialToggle', false) 
        self:AddCommandCap('RULEUCC_Move')
        local bpIntel = self:GetBlueprint().Intel
        self.CurrentRadar = bpIntel.RadarRadius
        self.RecloakCountdown = 0
        self.CloakManagerThread = self:ForkThread(self.CloakManager)
        self.DoneBeingBuilt = true
    end,

    Decloak = function(self)
        SWalkingLandUnit.Decloak(self)
        self:DisableUnitIntel('RadarStealth')
    end,

    Cloak = function(self)
        SWalkingLandUnit.Cloak(self)
        self:EnableUnitIntel('RadarStealth')
    end,

    OnMotionHorzEventChange = function(self, new, old)
        if new != 'Stopped' then
            ChangeState( self, self.VisibleState )
        end
        SWalkingLandUnit.OnMotionHorzEventChange(self, new, old)
    end,
    
    OnMotionHorzEventChange = function(self, new, old)
        OldUnit.OnMotionHorzEventChange(self, new, old)
        if old == 'Stopped' then
            self:RemoveToggleCap('RULEUTC_SpecialToggle')
        elseif new == 'Stopped' then
            self:AddToggleCap('RULEUTC_SpecialToggle')
        end
    end,
    EnableSpecialToggle = function(self)
        OldUnit.EnableSpecialToggle(self)
        if self.DoneBeingBuilt == true then
            if self.UndeployThread then
                self.UndeployThread:Destroy()
            end
            self:RemoveCommandCap('RULEUCC_Move')
            self:SetImmobile(true)
            self.DeployThread = self:ForkThread(self.Deploy)
        end
    end,
    DisableSpecialToggle = function(self)
        OldUnit.DisableSpecialToggle(self)
        if self.DoneBeingBuilt == true then
            if self.DeployThread then
                self.DeployThread:Destroy()
            end
            self.UndeployThread = self:ForkThread(self.Undeploy)
        end
    end,
    Undeploy = function(self)
        local bpIntel = self:GetBlueprint().Intel
        local MaxRadar = self.CurrentRadar
        local MinRadar = bpIntel.RadarRadius
        local StepSize = 0.25
        local RadarStepSize = (bpIntel.RadarRadiusBoost - bpIntel.RadarRadius) / 20
        while self.CurrentRadar > MinRadar do
            self.CurrentRadar = self.CurrentRadar - RadarStepSize
            self:SetIntelRadius('Radar', self.CurrentRadar)
            WaitSeconds(StepSize)
        end
        self:SetIntelRadius('Radar', MinRadar)
        self:AddCommandCap('RULEUCC_Move')
        self:SetImmobile(false)
    end,
    Deploy = function(self)
        local bpIntel = self:GetBlueprint().Intel
        local MaxRadar = bpIntel.RadarRadiusBoost
        local MinRadar = self.CurrentRadar
        local StepSize = 0.25
        local RadarStepSize = (bpIntel.RadarRadiusBoost - bpIntel.RadarRadius) / 20
        while self.CurrentRadar < MaxRadar do
            self.CurrentRadar = self.CurrentRadar + RadarStepSize
            self:SetIntelRadius('Radar', self.CurrentRadar)
            WaitSeconds(StepSize)
        end
        self:SetIntelRadius('Radar', MaxRadar)
    end,
}   
TypeClass = XSL0101
end
